/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data;

import ssg.tools.easyconfig.data.types.ISettingType;
import ssg.tools.easyconfig.data.types.SettingTypeFactory;

/**
 * Setting is generally a string value located at source. Value may be identified (found)
 * using arbitrary "locator". This locator may be used differently depending on
 * source type.
 *
 * Setting has default, loaded, and current value. Loaded is value found in source
 * while value is value to be in source. If these two differ then source must be
 * updated.
 *
 * Setting may fail when reading, validating new value, and applying value. The
 * 3 vars allow registration for further analysis of those problems.
 *
 * 
 * 
 * @author ssg
 */
public class NSetting {

    public static final long O_READONLY = 0x0001;
    public static final long O_MANDATORY = 0x0002;
    private NSettingSource source;
    private String group = "";
    private String name;
    private String locator;
    private String defaultValue;
    private String loadedValue;
    private String value;
    private long options;
    private String type = "string";
    private Throwable validationError;
    private Throwable loadError;
    private Throwable saveError;
    private NSetting parent;

    public NSetting(NSettingSource source,
            String name,
            String defaultValue,
            String loadedValue,
            String value, String type) {
        this.name = name;
        this.defaultValue = defaultValue;
        this.loadedValue = loadedValue;
        this.value = value;
        if (type != null) {
            this.type = type;
        }
        setSource(source);
    }

    public NSetting(NSettingSource source,
            String name,
            String defaultValue,
            String loadedValue,
            String value) {
        this.name = name;
        this.defaultValue = defaultValue;
        this.loadedValue = loadedValue;
        this.value = value;
        setSource(source);
    }

    public NSetting(NSettingSource source,
            String name,
            String defaultValue) {
        this.name = name;
        this.defaultValue = defaultValue;
        setSource(source);
    }

    public NSetting(NSettingSource source,
            String name,
            String defaultValue, String locator) {
        this.name = name;
        this.defaultValue = defaultValue;
        this.locator = locator;
        setSource(source);
    }

    @Override
    public String toString() {
        return name + " [" + type + ", " + locator + "] = " + value;
    }

    /**
     * Sets validation error if value does not pass validation.
     * 
     * @param value
     * @return 
     */
    public boolean validate(String value) {
        try {
            ISettingType nt = SettingTypeFactory.getInstance().getType(getType(), null);
            if (nt == null) {
                setValidationError(new NValidationException("No type is associated with setting " + this));
                System.err.println("Validation error for " + this + ": " + getValidationError());
                return false;
            }
            String v2 = nt.validate(this, value);
            setValidationError(null);
            // TODO: set successfully validated value?
            return true;
        } catch (NValidationException nvex) {
            System.err.println("Validation error for " + this + ": " + nvex);
            setValidationError(nvex);
            return false;
        }
    }

    /**
     * @return the source
     */
    public NSettingSource getSource() {
        return source;
    }

    /**
     * @param source the source to set
     */
    public void setSource(NSettingSource source) {
        if (this.source != null && !this.source.equals(source)) {
            NSettingSource old = this.source;
            this.source = null;
            old.removeSetting(this);
        }
        this.source = source;
        if (source != null) {
            this.source.addSetting(this);
        }
    }

    /**
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * @param name the name to set
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return the defaultValue
     */
    public String getDefaultValue() {
        return defaultValue;
    }

    /**
     * @param defaultValue the defaultValue to set
     */
    public void setDefaultValue(String defaultValue) {
        this.defaultValue = defaultValue;
    }

    /**
     * @return the loadedValue
     */
    public String getLoadedValue() {
        return loadedValue;
    }

    /**
     * @param loadedValue the loadedValue to set
     */
    public void setLoadedValue(String loadedValue) {
        this.loadedValue = loadedValue;
    }

    /**
     * @return the value
     */
    public String getValue() {
        return value;
    }

    /**
     * @param value the value to set
     */
    public void setValue(String value) {
        this.value = value;
    }

    /**
     * @return the validationError
     */
    public Throwable getValidationError() {
        return validationError;
    }

    /**
     * @param validationError the validationError to set
     */
    public void setValidationError(Throwable validationError) {
        this.validationError = validationError;
    }

    /**
     * @return the loadError
     */
    public Throwable getLoadError() {
        return loadError;
    }

    /**
     * @param loadError the loadError to set
     */
    public void setLoadError(Throwable loadError) {
        this.loadError = loadError;
    }

    /**
     * @return the saveError
     */
    public Throwable getSaveError() {
        return saveError;
    }

    /**
     * @param saveError the saveError to set
     */
    public void setSaveError(Throwable saveError) {
        this.saveError = saveError;
    }

    /**
     * @return the locator
     */
    public String getLocator() {
        return locator;
    }

    /**
     * @param locator the locator to set
     */
    public void setLocator(String locator) {
        this.locator = locator;
    }

    /**
     * @return the options
     */
    public long getOptions() {
        return options;
    }

    /**
     * @param options the options to set
     */
    public void setOptions(long options) {
        this.options = options;
    }

    /**
     * @return the type
     */
    public String getType() {
        return type;
    }

    /**
     * @param type the type to set
     */
    public void setType(String type) {
        if (type == null) {
            type = "string";
        }
        this.type = type;
    }

    ////////////////////////////////////////////////////////////// utilities
    public boolean isModified() {
        return isDifferent(loadedValue);
    }

    public boolean isDifferent(String value2) {
        return (value == null && value2 != null)
                || (value != null && value2 == null)
                || value != null && !value.equals(value2);
    }

    public boolean isReadOnly() {
        return (options & O_READONLY) == O_READONLY;
    }

    public boolean isMandatory() {
        return (options & O_MANDATORY) == O_MANDATORY;
    }

    /////////////////////////////////////////////////////////// identity
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final NSetting other = (NSetting) obj;
        if (this.source != other.source && (this.source == null || !this.source.equals(other.source))) {
            return false;
        }
        if ((this.name == null) ? (other.name != null) : !this.name.equals(other.name)) {
            return false;
        }
        if ((this.locator == null) ? (other.locator != null) : !this.locator.equals(other.locator)) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 5;
        hash = 43 * hash + (this.source != null ? this.source.hashCode() : 0);
        hash = 43 * hash + (this.name != null ? this.name.hashCode() : 0);
        hash = 43 * hash + (this.locator != null ? this.locator.hashCode() : 0);
        return hash;
    }

    /**
     * @return the group
     */
    public String getGroup() {
        return group;
    }

    /**
     * @param group the group to set
     */
    public void setGroup(String group) {
        this.group = group;
    }

    /**
     * @return the parent
     */
    public NSetting getParent() {
        return parent;
    }

    /**
     * @param parent the parent to set
     */
    public void setParent(NSetting parent) {
        this.parent = parent;
    }

}
