/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data;

import java.util.ArrayList;
import java.util.List;

/**
 * Sessing source defines source path, type, and state.
 * It owns set of settings accessible via source.
 * 
 * Setting source may be loaded (i.e. initialized from actual source)
 * and applied (i.e. modified settings are stored to source).
 * To check if there's setting that is modified use isModified() method.
 * 
 * Setting sources are usually initialized via setting source factory that uses
 * instance object to check if path/path object is supported by the source and
 * creates proper instance with newInstance(<path>) method.
 * 
 * Source is identified by type and path.
 * 
 * @author ssg
 */
public abstract class NSettingSource {

    /**
     * @return the loadedAt
     */
    public long getLoadedAt() {
        return loadedAt;
    }

    /**
     * @param loadedAt the loadedAt to set
     */
    public void setLoadedAt(long loadedAt) {
        this.loadedAt = loadedAt;
    }

    /**
     * @return the appliedAt
     */
    public long getAppliedAt() {
        return appliedAt;
    }

    /**
     * @param appliedAt the appliedAt to set
     */
    public void setAppliedAt(long appliedAt) {
        this.appliedAt = appliedAt;
    }
    private String path;
    private String type;
    private List<NSetting> settings = new ArrayList<NSetting>();
    private long loadedAt;
    private long appliedAt;

    public NSettingSource(String path, String type) {
        this.path = path;
        this.type = type;
    }

    public NSettingSource(String path) {
        this.path = path;
    }

    public NSettingSource addSetting(NSetting setting) {
        if (!settings.contains(setting)) {
            settings.add(setting);
            setting.setSource(this);
        }
        return this;
    }

    public NSettingSource removeSetting(NSetting setting) {
        if (settings.contains(setting)) {
            settings.remove(setting);
            setting.setSource(null);
        }
        return this;
    }

    ///////////////////////////////////////////////////////// properties
    /**
     * @return the path
     */
    public String getPath() {
        return (path!=null) ? path : "";
    }

    /**
     * @param path the path to set
     */
    public void setPath(String path) {
        this.path = path;
    }

    /**
     * @return the type
     */
    public String getType() {
        return type;
    }

    /**
     * @param type the type to set
     */
    public void setType(String type) {
        this.type = type;
    }

    /**
     * @return the settings
     */
    public List<NSetting> getSettings() {
        return settings;
    }

    /**
     * @param settings the settings to set
     */
    public void setSettings(List<NSetting> settings) {
        this.settings = settings;
    }

    @Override
    public String toString() {
        return getPath();
    }

    public boolean isModified() {
        if (settings != null) {
            for (NSetting ns : settings) {
                if (ns.isModified()) {
                    return true;
                }
            }
        }
        return false;
    }

    public abstract NSettingSource instanceFor(String path);

    ///////////////////////////////////////////////////////// identity
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final NSettingSource other = (NSettingSource) obj;
        if ((this.path == null) ? (other.path != null) : !this.path.equals(other.path)) {
            return false;
        }
        if (this.type != other.type) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 53 * hash + (this.path != null ? this.path.hashCode() : 0);
        hash = 53 * hash + (this.type != null ? this.type.hashCode() : 0);
        return hash;
    }
}
