/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler;

import ssg.tools.easyconfig.data.APPLY_STRATEGY;
import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.sources.NSettingSourceFactory;
import ssg.tools.easyconfig.data.sources.PropertiesSettingSource;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Basic functionality similar accross actual handlers.
 * 
 * @author ssg
 */
public abstract class BaseConfigurationHandler implements IConfigurationHandler {

    List<NSettingSource> settings = new ArrayList<NSettingSource>();
    private Map<NSetting, List<NSetting>> syncSettings = new HashMap<NSetting, List<NSetting>>();
    private Map<String, List<NSetting>> settingGroups = new LinkedHashMap<String, List<NSetting>>();
    protected String templateEncoding = "UTF-8";
    private long lastModified;
    private long lastLoaded;
    private APPLY_STRATEGY applyStrategy = APPLY_STRATEGY.backup;

    @Override
    public List<NSettingSource> getSettings() {
        ArrayList<NSettingSource> result = new ArrayList<NSettingSource>(settings.size());
        result.addAll(settings);
        return result;
    }

    @Override
    public void setSettings(List<NSettingSource> newSettings, boolean replace) {
        settingGroups.clear();
        if (replace) {
            settings.clear();
        }
        for (NSettingSource nss : newSettings) {
            if (!settings.contains(nss)) {
                settings.add(nss);
            }
        }
        setLastModified(System.currentTimeMillis());
    }

    @Override
    public List<String> getGroups() {
        List<String> result = new ArrayList<String>();
        if (settings != null && !settings.isEmpty() && settingGroups.isEmpty()) {
            initSettingGroups();
        }
        result.addAll(settingGroups.keySet());
        return result;
    }

    @Override
    public List<NSetting> getGroup(String group) {
        return settingGroups.get(group);
    }

    @Override
    public void initSettingGroups() {
        settingGroups.clear();
        if (settings == null) {
            return;
        }
        for (NSettingSource nss : settings) {
            if (nss == null) {
                continue;
            }
            for (NSetting ns : nss.getSettings()) {
                if (ns == null) {
                    continue;
                }
                String group = ns.getGroup();
                if (group == null || group.trim().isEmpty()) {
                    group = "Default";
                }
                List<NSetting> items = settingGroups.get(group);
                if (items == null) {
                    items = new ArrayList<NSetting>();
                    settingGroups.put(group, items);
                }
                items.add(ns);
            }
        }
    }

    @Override
    public void synchronizeSettings(NSetting setting1, NSetting setting2, boolean enable) {
        // 1 -> 2
        List<NSetting> lst = syncSettings.get(setting1);
        if (lst == null && enable) {
            lst = new ArrayList<NSetting>();
            syncSettings.put(setting1, lst);
        }
        if (lst != null && setting2 != null) {
            if (enable) {
                if (lst.contains(setting2)) {
                    lst.remove(setting2);
                }
            } else {
                if (!lst.contains(setting2)) {
                    lst.add(setting2);
                }
            }
        }
        // 2 -> 1
        lst = syncSettings.get(setting2);
        if (lst == null && enable) {
            lst = new ArrayList<NSetting>();
            syncSettings.put(setting2, lst);
        }
        if (lst != null && setting1 != null) {
            if (enable) {
                if (lst.contains(setting1)) {
                    lst.remove(setting1);
                }
            } else {
                if (!lst.contains(setting1)) {
                    lst.add(setting1);
                }
            }
        }
    }

    @Override
    public boolean isModified() {
        if (settings != null) {
            for (NSettingSource nss : settings) {
                if (nss.isModified()) {
                    return true;
                }
            }
        }
        return false;
    }
    static final String TMPL_S_COMPATIBILITY = "*Compatibility*";

    public void saveConfigurationTemplate(List<NSettingSource> settings, OutputStream os, String compatibility) throws IOException {
        System.out.println("Writing template to " + os);
        Writer wr = new OutputStreamWriter(new BufferedOutputStream(os), templateEncoding);
        wr.write("\n[" + TMPL_S_COMPATIBILITY + "]\n");
        wr.write(compatibility + "\n");
        for (NSettingSource nss : settings) {
            wr.write("\n[" + nss.getType() + ", " + nss.getPath() + "]\n");
            for (NSetting ns : nss.getSettings()) {
                wr.write("name=" + ns.getName() + "\n");
                wr.write("group=" + ns.getGroup() + "\n");
                wr.write("locator=" + ns.getLocator() + "\n");
                wr.write("default=" + ns.getDefaultValue() + "\n");
                wr.write("options=0x" + Long.toHexString(ns.getOptions()) + "\n");
                wr.write("type=" + ns.getType() + "\n\n");
                wr.write("value=" + ns.getValue() + "\n\n");
            }
        }
        wr.close();
    }

    public List<NSettingSource> loadConfigurationTemplate(InputStream is, String compatibility) throws IOException {
        List<NSettingSource> ss = new ArrayList<NSettingSource>();
        BufferedInputStream bis = new BufferedInputStream(is);
        if (!loadConfigurationTemplateCustom(bis, ss)) {
            loadConfigurationTemplateAsText(bis, ss);
        }
        return ss;
    }

    public boolean loadConfigurationTemplateCustom(BufferedInputStream bis, List<NSettingSource> ss) {
        return false;
    }

    public boolean loadConfigurationTemplateAsText(InputStream is, List<NSettingSource> ss) throws IOException {
        LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is, templateEncoding));
        String s = null;
        String templateCompatibility = null;
        NSettingSource nss = null;
        NSetting ns = null;
        while ((s = lnr.readLine()) != null) {
            s = s.trim();
            if (s.startsWith("#") || s.startsWith("//") || s.isEmpty()) {
                continue;
            }
            if (s.startsWith("[")) {
                s = s.substring(1);
                if (s.endsWith("]")) {
                    s = s.substring(0, s.length() - 1);
                }
                s = s.trim();
                if (s.equals(TMPL_S_COMPATIBILITY)) {
                    nss = null;
                    ns = new NSetting(nss, TMPL_S_COMPATIBILITY, TMPL_S_COMPATIBILITY, TMPL_S_COMPATIBILITY);
                    ns.setValue("");
                } else {
                    int idx = s.indexOf(",");
                    String type = PropertiesSettingSource.SOURCE_TYPE;
                    if (idx != -1) {
                        try {
                            type = s.substring(0, idx).trim();
                        } catch (Throwable th) {
                            th.printStackTrace();
                        }
                        s = s.substring(idx + 1).trim();
                    }
                    nss = NSettingSourceFactory.getInstance().newSource(type, s);
                    ss.add(nss);
                    if (ns != null && ns.getSource() == null) {
                        templateCompatibility = ns.getValue();
                    }
                    ns = null;
                }
            } else {
                if (ns != null && ns.getSource() == null) {
                    // compatibility values
                    ns.setValue(ns.getValue() + s + "\n");
                } else {
                    int idx = s.indexOf("=");
                    String key = s.substring(0, idx);
                    String value = s.substring(idx + 1);
                    if ("name".equals(key)) {
                        ns = new NSetting(nss, value, null, null);
                    } else if ("group".equals(key)) {
                        ns.setGroup(value);
                    } else if ("locator".equals(key)) {
                        ns.setLocator(value);
                    } else if ("value".equals(key)) {
                        ns.setValue(value);
                    } else if ("default".equals(key)) {
                        ns.setDefaultValue(value);
                    } else if ("type".equals(key)) {
                        ns.setType(value);
                    } else if ("options".equals(key)) {
                        try {
                            if (value.toLowerCase().startsWith("0x")) {
                                ns.setOptions(Long.parseLong(value.substring(2), 16));
                            } else {
                                ns.setOptions(Long.parseLong(value));
                            }
                        } catch (Throwable th) {
                            th.printStackTrace();
                        }
                    }
                }
            }
        }
        lnr.close();
        return true;
    }

    /**
     * @return the lastModified
     */
    @Override
    public long getLastModified() {
        return lastModified;
    }

    /**
     * @param lastModified the lastModified to set
     */
    public void setLastModified(long lastModified) {
        this.lastModified = lastModified;
    }

    /**
     * @return the lastLoaded
     */
    @Override
    public long getLastLoaded() {
        return lastLoaded;
    }

    /**
     * @param lastLoaded the lastLoaded to set
     */
    public void setLastLoaded(long lastLoaded) {
        this.lastLoaded = lastLoaded;
    }

    /**
     * @return the applyStrategy
     */
    @Override
    public APPLY_STRATEGY getApplyStrategy() {
        return applyStrategy;
    }

    /**
     * @param applyStrategy the applyStrategy to set
     */
    @Override
    public void setApplyStrategy(APPLY_STRATEGY applyStrategy) {
        this.applyStrategy = applyStrategy;
    }
}
