/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler;

import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.handler.support.ISourceHandler;
import ssg.tools.easyconfig.data.handler.support.SourceHandlerFactory;
import ssg.tools.easyconfig.data.sources.IReadableSource;
import ssg.tools.easyconfig.data.sources.IWritableSource;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import ssg.tools.easyconfig.data.handler.support.INestedSourceHandler;
import ssg.tools.easyconfig.utils.Tree;

/**
 * Primary configuration implementation providing support for independent and 
 * nested source handlers.
 * 
 * @author ssg
 */
public class CompositeConfigurationHandler extends BaseConfigurationHandler {

    URI uri;
    Map<ISourceHandler, List<NSettingSource>> handlers = new LinkedHashMap<ISourceHandler, List<NSettingSource>>();
    Set<Tree<ISourceHandler>> dependencies = new HashSet<Tree<ISourceHandler>>();

    public CompositeConfigurationHandler(URI uri) throws IOException {
        this.uri = uri;
    }

    public CompositeConfigurationHandler(URL url) throws IOException {
        try {
            this.uri = url.toURI();
        } catch (URISyntaxException usex) {
            throw new IOException("Failed to convert URL '" + url + "' to URI: " + usex, usex);
        }
    }

    public CompositeConfigurationHandler(File file) throws IOException {
        this.uri = file.toURI();
    }

    public CompositeConfigurationHandler(URI uri, List<NSettingSource> settings) throws IOException {
        this.uri = uri;
        setSettings(settings, true);
    }

    public CompositeConfigurationHandler(URL url, List<NSettingSource> settings) throws IOException {
        try {
            this.uri = url.toURI();
        } catch (URISyntaxException usex) {
            throw new IOException("Failed to convert URL '" + url + "' to URI: " + usex, usex);
        }
        setSettings(settings, true);
    }

    public CompositeConfigurationHandler(File file, List<NSettingSource> settings) throws IOException {
        this.uri = file.toURI();
        setSettings(settings, true);
    }

    @Override
    public void setSettings(List<NSettingSource> newSettings, boolean replace) {
        super.setSettings(newSettings, replace);
        init();
    }

    @Override
    public String toString() {
        return getLocation() + " (" + handlers.size() + " handlers)";
    }

    public String getLocation() {
        return (uri != null) ? uri.toString() : "unknown";
    }

    public void initSettingsList() {
        //throw new UnsupportedOperationException("Not supported yet.");
    }

    public void loadSettings() throws IOException {
        Set<ISourceHandler> processed = new HashSet<ISourceHandler>();
        for (ISourceHandler sh : handlers.keySet()) {
            loadSettings(sh, processed);
        }
        setLastLoaded(System.currentTimeMillis());
    }

    public void loadSettings(ISourceHandler sh, Set<ISourceHandler> processed) throws IOException {
        if (processed.contains(sh)) {
            return;
        }
        processed.add(sh);
        List<NSettingSource> sources = handlers.get(sh);
        Map<String, NSettingSource> mappedNSS = new HashMap<String, NSettingSource>();
        {
            INestedSourceHandler nsh = (sh instanceof INestedSourceHandler) ? (INestedSourceHandler) sh : null;
            for (NSettingSource nss : sources) {
                String p = nss.getPath();
                if (p.contains("!") && nsh != null) {
                    String pp = p.substring(0, p.lastIndexOf("!"));
                    if (nsh.getUri().toString().endsWith(pp)) {
                        p = p.substring(p.lastIndexOf("!") + 1);
                    }
                }
                mappedNSS.put(p, nss);
            }
        }
        int processedSettings = 0;
        Tree<ISourceHandler> tt = findTT(sh);
        sh.open(true, getApplyStrategy());
        while (sh.hasNext() && processedSettings < mappedNSS.size()) {
            sh.next();
            NSettingSource nss = mappedNSS.get(sh.getName());
            if (nss != null && nss instanceof IReadableSource) {
                if (((IReadableSource) nss).load(sh.getInput())) {
                    processedSettings++;
                }
            } else {
                // check if there's handler interested in current sh item...
                try {
                    URI luri = new URI(sh.getUri().toString() + "/" + sh.getName());
                    for (ISourceHandler nsh : handlers.keySet()) {
                        if (processed.contains(nsh)) {
                            continue;
                        }
                        if (tt != null && tt.path(nsh).contains(sh)) {
                            if (nsh.getUri().equals(luri)) {
                                if (nsh instanceof INestedSourceHandler) {
//                                Tree<ISourceHandler> tp = findTT(sh);
//                                if (tp == null) {
//                                    tp = new Tree<ISourceHandler>(sh);
//                                    dependencies.add(tp);
//                                }
//                                tp.addChild(nsh);
                                    INestedSourceHandler nnsh = (INestedSourceHandler) nsh;
                                    nnsh.setParentSource(sh);
                                    loadSettings(nsh, processed);
                                    nnsh.setParentSource(null);
                                }
                            }
                        }
                    }
                } catch (URISyntaxException usex) {
                    usex.printStackTrace();
                }
            }
        }
        sh.close();
    }

    public void evaluateDependencies() {
        for (ISourceHandler sh1 : handlers.keySet()) {
            for (ISourceHandler sh2 : handlers.keySet()) {
                if (sh1.equals(sh2)) {
                    continue;
                }
                String s1 = sh1.getUri().toString();
                String s2 = sh2.getUri().toString();
                if (s2.startsWith(s1)) {
                    Tree<ISourceHandler> t2 = findTT(sh2);
                    if (t2 == null) {
                        Tree<ISourceHandler> t1 = findTT(sh1);
                        if (t1 == null) {
                            t1 = new Tree<ISourceHandler>(sh1);
                            dependencies.add(t1);
                        }
                        t1.addChild(sh2);
                    }
                }
            }
        }
    }

    Tree<ISourceHandler> findTT(ISourceHandler sh) {
        for (Tree<ISourceHandler> tt : dependencies) {
            Tree<ISourceHandler> tt2 = tt.find(sh, 1000);
            if (tt2 != null) {
                return tt2;
            }
        }
        return null;
    }

    public void saveSettings() throws IOException {
        Set<ISourceHandler> processed = new HashSet<ISourceHandler>();
        for (ISourceHandler sh : handlers.keySet()) {
            saveSettings(sh, processed);
        }
    }

    public void saveSettings(ISourceHandler sh, Set<ISourceHandler> processed) throws IOException {
        List<NSettingSource> sources = handlers.get(sh);
        if (processed.contains(sh)) {
            return;
        }
        processed.add(sh);
        //System.out.println("SH: " + sh.getName() + ", " + sh.getUri());
        Map<String, NSettingSource> mappedNSS = new HashMap<String, NSettingSource>();
        {
            INestedSourceHandler nsh = (sh instanceof INestedSourceHandler) ? (INestedSourceHandler) sh : null;
            for (NSettingSource nss : sources) {
                String p = nss.getPath();
                if (p.contains("!") && nsh != null) {
                    String pp = p.substring(0, p.lastIndexOf("!"));
                    if (nsh.getUri().toString().endsWith(pp)) {
                        p = p.substring(p.lastIndexOf("!") + 1);
                    }
                }
                mappedNSS.put(p, nss);
            }
        }
        int processedSettings = 0;
        Tree<ISourceHandler> tt = findTT(sh);
        sh.open(false, getApplyStrategy());
        while (sh.hasNext() && processedSettings < mappedNSS.size()) {
            sh.next();
            NSettingSource nss = mappedNSS.get(sh.getName());
            //System.out.println("About to save " + sh.getName());
            if (nss != null && nss.isModified() && nss instanceof IWritableSource) {
                //System.out.println("loadSettings:" + file + "!" + sh.getName() + ", " + nss);
                processedSettings++;
                sh.prepareForWrite(true);
                System.out.println("Save with processing " + sh.getName());
                if (!((IWritableSource) nss).save(sh.getInput(), sh.getOutput())) {
                    //sr.copyInputToOutput();
                    throw new IOException("Failed to apply settings to " + nss);
                }
                continue;
            } else if (nss == null) {
                // check if there's handler interested in current sh item...
                try {
                    URI luri = new URI(sh.getUri().toString() + "/" + sh.getName());
                    boolean isProcessed = false;
                    for (ISourceHandler nsh : handlers.keySet()) {
                        if (processed.contains(nsh)) {
                            continue;
                        }
                        if (tt != null && tt.path(nsh).contains(sh)) {
                            if (nsh.getUri().equals(luri)) {
                                if (nsh instanceof INestedSourceHandler) {
                                    sh.prepareForWrite(true);
                                    INestedSourceHandler nnsh = (INestedSourceHandler) nsh;
                                    nnsh.setParentSource(sh);
                                    //System.out.println("Save nested " + sh.getName());
                                    saveSettings(nsh, processed);
                                    nnsh.setParentSource(null);
                                    isProcessed = true;
                                    break;
                                }
                            }
                        }
                    }
                    if (isProcessed) {
                        continue;
                    }
                } catch (URISyntaxException usex) {
                    usex.printStackTrace();
                }
            }
            // copy input to output
            //System.out.println("Save as copy " + sh.getName());
            sh.copyInputToOutput();
        }
        sh.close();
    }

    ////////////////////////////////////////////////////////// utilities
    /**
     * Prepares list of source handlers and related setting sources.
     * Initializes sources dependencies by calling <method>evaluateDependencies</method>.
     * 
     */
    public void init() {
        handlers.clear();
        for (NSettingSource nss : getSettings()) {
            boolean resolved = false;
            for (ISourceHandler sh : handlers.keySet()) {
                try {
                    if (sh.canRead(nss)) {
                        handlers.get(sh).add(nss);
                        resolved = true;
                        break;
                    }
                } catch (IOException ioex) {
                }
            }
            if (resolved) {
                continue;
            }
            ISourceHandler nsh = SourceHandlerFactory.getInstance().newSourceHandler(uri, nss);
            if (nsh != null) {
                List<NSettingSource> sources = new ArrayList<NSettingSource>();
                sources.add(nss);
                handlers.put(nsh, sources);
            }
        }
        for (ISourceHandler sh : handlers.keySet()) {
            try {
                sh.init(handlers.get(sh));
            } catch (IOException ioex) {
                ioex.printStackTrace();
            }
        }

        evaluateDependencies();
        System.out.println("Discovered dependencies: " + dependencies);
    }

    public Collection<ISourceHandler> getSourceHandlers() {
        return this.handlers.keySet();
    }

    public URI getUri() {
        return uri;
    }
}
