/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler;

import ssg.tools.easyconfig.data.handler.support.SourceHandlerFactory;
import ssg.tools.easyconfig.data.handler.support.ISourceHandler;
import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.sources.IReadableSource;
import ssg.tools.easyconfig.data.sources.IWritableSource;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Configuration handler based on file system.
 * 
 * @author ssg
 */
public class FileConfigurationHandler extends BaseConfigurationHandler {

    File configFile;
    URL configURL;
    URI configURI;
    ISourceHandler sr;

    public FileConfigurationHandler(File file) throws IOException {
        configFile = file;
        configURI = configFile.toURI();
        sr = SourceHandlerFactory.getInstance().newSourceHandler(configURI);
        try {
            loadSettings();
        } catch (IOException ioex) {
            System.err.println("Error while loading settings: " + ioex);
            throw ioex;
        }
    }

    public FileConfigurationHandler(File file, List<NSettingSource> settings) throws IOException {
        configFile = file;
        configURI = configFile.toURI();
        setSettings(settings, true);
        sr = SourceHandlerFactory.getInstance().newSourceHandler(configURI);
        try {
            loadSettings();
        } catch (IOException ioex) {
            System.err.println("Error while loading settings: " + ioex);
            throw ioex;
        }
    }

    public FileConfigurationHandler(URL url) {
        configURL = url;
        try {
            configURI = url.toURI();
        } catch (URISyntaxException usex) {
            usex.printStackTrace();
        }
    }

    @Override
    public String toString() {
        return configFile.getName() + " [" + ((sr != null) ? sr.getClass().getSimpleName() : "unknown") + "]";
    }

    @Override
    public String getLocation() {
        return (configFile != null) ? configFile.getAbsolutePath() : "" + configURL;
    }

    @Override
    public void loadSettings() throws IOException {
        Map<String, NSettingSource> mappedNSS = new HashMap<String, NSettingSource>();
        for (NSettingSource nss : settings) {
            mappedNSS.put(nss.getPath(), nss);
        }
        int processedSettings = 0;
        sr.open(true,getApplyStrategy());
        while (sr.hasNext() && processedSettings < mappedNSS.size()) {
            sr.next();
            NSettingSource nss = mappedNSS.get(sr.getName());
            if (nss != null && nss instanceof IReadableSource) {
                if (((IReadableSource) nss).load(sr.getInput())) {
                    processedSettings++;
                }
            }
        }
        sr.close();
        setLastLoaded(System.currentTimeMillis());
    }

    @Override
    public void saveSettings() throws IOException {
        Map<String, NSettingSource> mappedNSS = new HashMap<String, NSettingSource>();
        for (NSettingSource nss : settings) {
            mappedNSS.put(nss.getPath(), nss);
        }
        int processedSettings = 0;
        sr.open(false,getApplyStrategy());
        while (sr.hasNext() && processedSettings < mappedNSS.size()) {
            sr.next();
            NSettingSource nss = mappedNSS.get(sr.getName());
            if (nss != null && nss.isModified() && nss instanceof IWritableSource) {
                processedSettings++;
                sr.prepareForWrite(true);
                if (!((IWritableSource) nss).save(sr.getInput(), sr.getOutput())) {
                    throw new IOException("Failed to apply settings to " + nss);
                }
            } else {
                // copy input to output
                sr.copyInputToOutput();
            }
        }
        sr.close();
    }

    /**
     * No settings by default
     */
    public void initSettingsList() {
    }

    public URI getUri() {
        return configURI;
    }
}
