/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler;

import ssg.tools.easyconfig.data.APPLY_STRATEGY;
import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NSettingSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.util.List;

/**
 *
 * @author ssg
 */
public interface IConfigurationHandler {

    /**
     * String representation of data associated with handler. NOTE: actual
     * locations are handled via nested source hanfdler objects.
     *
     * @return
     */
    String getLocation();

    /**
     * Returns primary URI associated with handler.
     *
     * @return
     */
    URI getUri();

    /**
     * List of managed sources with settings.
     *
     * @return
     */
    List<NSettingSource> getSettings();

    /**
     * List of settings groups within setting sources. Setting groups peovide
     * LOGICAL subsets of settings while setting sources - physical.
     *
     * @return
     */
    List<String> getGroups();

    /**
     * Returns list of settings for logical group.
     *
     * @param group
     * @return
     */
    List<NSetting> getGroup(String group);

    /**
     * May be used to initialize default settings.
     */
    void initSettingsList();

    /**
     * Initializes logical groups of settings. Usually is called when settings
     * are assigned.
     */
    void initSettingGroups();

    /**
     * Returns true if at least 1 setting is modified.
     *
     * @return
     */
    boolean isModified();

    /**
     * Loads settings from source handler(s). USed to initialize settings with
     * loaded value/value from appropriate source(s).
     *
     * @throws IOException
     */
    void loadSettings() throws IOException;

    /**
     * Save settings via source handlers. Source handlers may override or create
     * alternative output.
     *
     * @throws IOException
     */
    void saveSettings() throws IOException;

    /**
     * Assigns setting sources. If replace is true then existing settimgs are
     * removed prior to assignment of new ones.
     *
     * @param newSettings
     * @param replace
     */
    void setSettings(List<NSettingSource> newSettings, boolean replace);

    /**
     * Binds/unbinds seetings that are in sync. Used to enable single change for
     * settings that are placed in different locations.
     *
     * @param setting1
     * @param setting2
     * @param enable
     */
    void synchronizeSettings(NSetting setting1, NSetting setting2, boolean enable);

    /**
     * Loads settings from external source. Complimentary with <method>saveConfigurationTemplate</method>.
     *
     * @param is
     * @param compatibility
     * @return
     * @throws IOException
     */
    List<NSettingSource> loadConfigurationTemplate(InputStream is, String compatibility) throws IOException;

    /**
     * Save settings into text file. Complimentary with <method>loadConfigurationTemplate</method>.
     *
     * @param settings
     * @param configTemplate
     * @param compatibility
     * @throws IOException
     */
    void saveConfigurationTemplate(List<NSettingSource> settings, OutputStream os, String compatibility) throws IOException;

    /**
     * Returns latest modification timestamp.
     *
     * @return
     */
    long getLastModified();

    /**
     * Returns timestamp for last loading action.
     *
     * @return
     */
    long getLastLoaded();

    /**
     * Returns apply strategy in use.
     * 
     * @return 
     */
    APPLY_STRATEGY getApplyStrategy();

    /**
     * Change apply strategy to use.
     * 
     * @param newStrategy 
     */
    void setApplyStrategy(APPLY_STRATEGY newStrategy);
}
