/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;

/**
 * Generic file list oriented source handler: provides operations for
 * reading/writing changes. Normal operation sequence: open [hasNext next
 * getInput prepareForWrite getOutput copyInputToOutput] close
 *
 */
public abstract class AbstractFileStreamSourceHandler<T> implements ISourceHandler<InputStream, OutputStream> {
    //

    private APPLY_STRATEGY applyStrategy;
    List<T> files = new ArrayList<T>();
    // reader
    InputStream is;
    private int idx;
    // writer
    OutputStream os;
    URI uri;

    public AbstractFileStreamSourceHandler(URI uri) {
        this.uri = uri;
    }

    /**
     * Initializes source based on type. For read-only mode only input stream
     * can be requested and output stream is not accessible.
     *
     * @param readOnly
     * @throws IOException
     */
    public void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException {
        idx = -1;
        this.applyStrategy = (applyStrategy != null) ? applyStrategy : APPLY_STRATEGY.backup;
    }

    /**
     * Returns true if next source item is available. Must be called before 1st
     * item is used.
     *
     * @return
     */
    public boolean hasNext() {
        return files != null && idx < (files.size() - 1);
    }

    /**
     * Makes next item current. Closes any previous opened resources (files,
     * streams, etc.)
     *
     * @throws IOException
     */
    public void next() throws IOException {
        if (is != null) {
            try {
                is.close();
            } catch (IOException ioex) {
                ioex.printStackTrace();
            }
            is = null;
        }
        if (os != null) {
            os.close();
            applyStrategy();
            os = null;
        }
        if (hasNext()) {
            idx++;
        }
    }

    /**
     * Must be called before writing to output stream for current item. E.g.
     * prepares zip entry (see copyInputToOutput for details.
     *
     * @param modified
     * @throws IOException
     */
    public void prepareForWrite(boolean modified) throws IOException {
    }

    public void copyInputToOutput() throws IOException {
        prepareForWrite(false);
    }

    /**
     * Closes any open resources.
     *
     * @throws IOException
     */
    public void close() throws IOException {
        // to close any open resources...
        next();
        files.clear();
        idx = -1;
    }

    public void init(Object owner) throws IOException {
        // check if we need owner
    }

    public URI getUri() {
        return uri;
    }

    /**
     * Returns input stream for current item.
     *
     * @return
     * @throws IOException
     */
    public InputStream getInput() throws IOException {
        if (is == null) {
            is = openInputStream(current());
        }
        return is;
    }

    /**
     * Returns output stream for current item to store modified version.
     *
     * @return
     * @throws IOException
     */
    public <T extends OutputStream> T getOutput() throws IOException {
        if (os == null) {
            os = openOutputStream(current());
        }
        return (T) os;
    }

    public T current() {
        return (files != null && idx < files.size()) ? files.get(idx) : null;
    }

    public List<T> getFiles() {
        return files;
    }

    public abstract InputStream openInputStream(T file) throws IOException;

    public abstract OutputStream openOutputStream(T file) throws IOException;

    /**
     * Implement post-save actions based on strategy
     */
    public abstract void applyStrategy() throws IOException;

    /**
     * @return the applyStrategy
     */
    public APPLY_STRATEGY getApplyStrategy() {
        return applyStrategy;
    }
    
    @Override
    public String toString() {
        return getClass().getSimpleName()+"("+getUri()+")";
    }
}
