/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import ssg.tools.easyconfig.utils.IOUtils;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.util.List;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;

/**
 * Generic source handler: provides operations fro reading/writing changes.
 * Normal operation sequence: open [hasNext next getInput prepareForWrite
 * getOutput copyInputToOutput] close
 *
 * TODO: allow URL based folders: - ftp - webdav - ssh - svn
 *
 */
public class FolderStreamSourceHandler extends AbstractFileStreamSourceHandler<File> {

    String basePath;
    List<File> files;

    public FolderStreamSourceHandler() {
        super(null);
    }

    public FolderStreamSourceHandler(File file) {
        super(file.toURI());
        basePath = file.getAbsolutePath();
    }

    public FolderStreamSourceHandler(URI uri) {
        super(uri);
        if (uri != null) {
            basePath = (new File(uri)).getAbsolutePath();
        }
    }

    /**
     * Initializes source baed on type. For read-only mode only input stream can
     * be requested and output stream is not accessible.
     *
     * @param readOnly
     * @throws IOException
     */
    public void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException {
        super.open(readOnly, applyStrategy);
        IOUtils.listFiles(new File(getUri()), getFiles());
    }

    /**
     * Makes next item current. Closes any previous opened resources (files,
     * streams, etc.)
     *
     * @throws IOException
     */
    public void applyStrategy() throws IOException {
        File oldF = current();
        File newF = new File(oldF.getAbsolutePath() + ".tmp");
        File bakF = new File(oldF.getAbsolutePath() + ".bak");
        switch (getApplyStrategy()) {
            case replace:
                oldF.delete();
                newF.renameTo(oldF);
                break;
            case backup:
                if (bakF.exists()) {
                    int bakIdx = 0;
                    File bakF2 = new File(bakF.getAbsolutePath() + "." + bakIdx);
                    while (bakF2.exists()) {
                        bakIdx++;
                        bakF2 = new File(bakF.getAbsolutePath() + "." + bakIdx);
                    }
                    bakF = bakF2;
                }
                oldF.renameTo(bakF);
                newF.renameTo(oldF);
                break;
            case other:
                break;
            default:
                break;
        }
    }

    /**
     * Retruns current item name.
     *
     * @return
     */
    public String getName() {
        File f = current();
        return (f != null) ? f.getAbsolutePath().substring(basePath.length() + 1).replace('\\', '/') : null;
    }

    public boolean canRead(Object obj) throws IOException {
        File file = null;
        try {
            if (obj instanceof File) {
                file = (File) obj;
            } else if (obj instanceof URI) {
                file = ("file".equals(((URI) obj).getScheme())) ? new File((URI) obj) : null;
            } else if (obj instanceof URL) {
                file = new File(((URL) obj).toURI());
            } else if (obj instanceof String) {
                file = new File(((String) obj));
            }
        } catch (Throwable th) {
            // TODO: log info?
            th.printStackTrace();
        }
        return file != null && file.isDirectory();
    }

    @Override
    public InputStream openInputStream(File file) throws IOException {
        if (file != null) {
            return new BufferedInputStream(new FileInputStream(file));
        } else {
            return null;
        }
    }

    @Override
    public OutputStream openOutputStream(File file) throws IOException {
        if (file != null) {
            return new BufferedOutputStream(new FileOutputStream(new File(file.getAbsolutePath() + ".tmp")));
        } else {
            return null;
        }
    }

    public ISourceHandler<InputStream, OutputStream> instanceFor(URI uriIn, URI uriOut) {
        try {
            if (canRead(uriIn)) {
                return new FolderStreamSourceHandler(uriIn);
            }
        } catch (IOException ioex) {
        }
        return null;
    }
}
