/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import java.io.IOException;
import java.net.URI;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;

/**
 * Abstraction of source handler.
 * 
 * @author ssg
 */
public interface ISourceHandler<I, O> {

    void init(Object owner) throws IOException;

    /**
     * Closes any open resources.
     *
     * @throws IOException
     */
    void close() throws IOException;

    void copyInputToOutput() throws IOException;

    /**
     * Returns input stream for current item.
     *
     * @return
     * @throws IOException
     */
    I getInput() throws IOException;

    /**
     * Retruns current item name.
     *
     * @return
     */
    String getName();

    /**
     * Returns primary URI associated with handler.
     *
     * @return
     */
    URI getUri();

    /**
     * Returns output stream for current item to store modified version.
     *
     * @return
     * @throws IOException
     */
    <T extends O> T getOutput() throws IOException;

    /**
     * Returns true if next source item is available. Must be called before 1st
     * item is used.
     *
     * @return
     */
    boolean hasNext();

    /**
     * Makes next item current. Closes any previous opened resources (files,
     * streams, etc.)
     *
     * @throws IOException
     */
    void next() throws IOException;

    /**
     * Initializes source baed on type. For read-only mode only input stream can
     * be requested and output stream is not accessible.
     *
     * @param readOnly
     * @throws IOException
     */
    void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException;

    /**
     * Must be called before writing to output stream for current item. E.g.
     * prepares zip entry (see copyInputToOutput for details.
     *
     * @param modified
     * @throws IOException
     */
    void prepareForWrite(boolean modified) throws IOException;

    /**
     * Detect if provided object is supported by handler. NOTE: may have side
     * effects if used for InputStream
     *
     * @param obj
     * @return
     * @throws IOException
     */
    boolean canRead(Object obj) throws IOException;

    /**
     * Creates instance of source handker for specified URIs. uriOut is optional
     * and may be not supported.
     *
     * @param uri
     * @return
     */
    ISourceHandler<I, O> instanceFor(URI uriIn, URI uriOut);
}
