/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.handler.IConfigurationHandler;
import ssg.tools.easyconfig.utils.JDBCContext;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;

/**
 * Sample of non-stream source handler providing access to database on field level.
 * 
 * @author ssg
 */
public class JDBCSourceHandler implements ISourceHandler<JDBCContext, JDBCContext> {

    JDBCContext input;
    JDBCContext output;
    URI inputURI;
    URI outputURI;
    List<String> items = new ArrayList<String>();
    int idx = -1;

    public JDBCSourceHandler() throws IOException {
    }

    public JDBCSourceHandler(URI uri) throws IOException {
        this.inputURI = uri;
    }

    public JDBCSourceHandler(URI uriIn, URI uriOut) throws IOException {
        this.inputURI = uriIn;
        this.outputURI = uriOut;
    }

    public JDBCSourceHandler(String uri) throws IOException {
        try {
            this.inputURI = new URI(uri);
        } catch (URISyntaxException usex) {
            throw new IOException("Failed to create JDBC source handler for '" + uri + "': " + usex, usex);
        }
    }

    public void close() throws IOException {
        try {
            if (input != null) {
                input.disconnect();
            }
        } catch (SQLException sex) {
            throw new IOException("Failed to close connection: " + sex, sex);
        }
        try {
            if (output != null) {
                if (System.identityHashCode(input) != System.identityHashCode(output)) {
                    output.disconnect();
                }
            }
        } catch (SQLException sex) {
            throw new IOException("Failed to close output connection: " + sex, sex);
        }
        input = null;
        output = null;
    }

    /**
     * Not applicable: can copy only values defined in settings and here we can get if setting is not defined.
     * 
     * @throws IOException 
     */
    public void copyInputToOutput() throws IOException {
//        if (input != null && !input.equals(output)) {
//            for(String expr)
//            output.write(ns.getLocator(), input.read(ns.getLocator()));
//        }
    }

    public JDBCContext getInput() throws IOException {
        return input;
    }

    public String getName() {
        return (idx >= 0 && items != null && idx < items.size()) ? items.get(idx) : null;
    }

    public <T extends JDBCContext> T getOutput() throws IOException {
        return (T) output;
    }

    public boolean hasNext() {
        return items != null && idx < (items.size() - 1);
    }

    public void next() throws IOException {
        if (hasNext()) {
            idx++;
        }
    }

    public void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException {
        if (inputURI == null) {
            throw new IOException("No URI is defined for JDBC source handler.");
        }
        if (input == null) {
            input = new JDBCContext(inputURI);
        }
        if (outputURI != null && !outputURI.equals(inputURI)) {
            output = new JDBCContext(outputURI);
        } else {
            output = input;
        }
    }

    public void prepareForWrite(boolean modified) throws IOException {
        // TODO: nothing to prepare?
    }

    public boolean canRead(Object obj) throws IOException {
        if (obj == null) {
            return false;
        }
        String objURI = null;
        if (obj instanceof NSettingSource) {
            objURI = ((NSettingSource) obj).getPath();
        } else if (obj instanceof String) {
            objURI = (String) obj;
        } else if (obj instanceof URI) {
            objURI = ((URI) obj).toString();
        }
        if (inputURI != null) {
            return inputURI.toString().equals(objURI);
        } else {
            if (objURI != null) {
                return objURI.startsWith("jdbc:");
            } else {
                return false;
            }
        }
    }

    public void init(Object owner) throws IOException {
        // TODO: try to evaluate list of settings based on owner...
        List list = (owner instanceof List) ? (List) owner : Collections.singletonList(owner);
        for (Object obj : list) {
            if (obj instanceof IConfigurationHandler) {
                IConfigurationHandler ch = (IConfigurationHandler) obj;
                for (NSettingSource nss : ch.getSettings()) {
                    initSource(nss);
                }
            } else if (obj instanceof NSettingSource) {
                initSource((NSettingSource) obj);
            }
        }
        idx = -1;
    }

    public void initSource(NSettingSource nss) throws IOException {
        if (this.canRead(nss.getPath())) {
            items.add(nss.getPath());
        }
    }

    public URI getUri() {
        return inputURI;
    }

    public ISourceHandler<JDBCContext, JDBCContext> instanceFor(URI uriIn, URI uriOut) {
        try {
            if (canRead(uriIn)) {
                return new JDBCSourceHandler(uriIn, uriOut);
            }
        } catch (IOException ioex) {
        }
        return null;
    }
    
    @Override
    public String toString() {
        return getClass().getSimpleName()+"("+getUri()+")";
    }
}
