/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import java.io.IOException;
import java.net.URI;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;

/**
 * Allows execution of child ISS methods combined with parent. Parent ISS
 * provides input and output for the child while child provides processing of
 * elements.
 * 
 * Logic: ISS cycle is (open, hasNext, [next, getInput, getOutput,] close)
 * In wrapper on open parent is initialized and prepared (parent.next()).
 * Then child is initialized (able to use paren't getInput, getOutput)
 * 
 * Then normal calls hasNext, next, getInput, getOutput are passed to child.
 * 
 * Then child is closed and parent is closed.
 * 
 * As a result we iterate thru objects handled by child ISS but on super-data
 * provided by parent ISS.
 * 
 * Sample use case: ZIP file on SFTP source.
 *
 * @author ssg
 */
public class NestedSourceWrapper<I, O> implements ISourceHandler<I, O> {
    
    ISourceHandler<I, O> parent;
    ISourceHandler<I, O> child;
    
    public NestedSourceWrapper(ISourceHandler<I, O> parent, INestedSourceHandler<I, O> child) {
        this.parent = parent;
        this.child = child;
        child.setParentSource(parent);
    }
    
    public void init(Object owner) throws IOException {
        // TODO: link parent and child with each other ??
    }

    ///////////////////////////////////////////////////////////////////// parent
    public URI getUri() {
        return parent.getUri();
    }
    
    public boolean canRead(Object obj) throws IOException {
        return parent.canRead(obj);
    }
    ////////////////////////////////////////////////// parent + child operations

    public void close() throws IOException {
        child.close();
        parent.close();
    }
    
    public void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException {
        parent.open(readOnly, applyStrategy);
        parent.next();
        child.open(readOnly, applyStrategy);
    }

    /////////////////////////////////////////////////////////// child operations
    public void copyInputToOutput() throws IOException {
        child.copyInputToOutput();
    }
    
    public I getInput() throws IOException {
        return child.getInput();
    }
    
    public String getName() {
        return child.getName();
    }
    
    public <T extends O> T getOutput() throws IOException {
        return child.getOutput();
    }
    
    public boolean hasNext() {
        return child.hasNext();
    }
    
    public void next() throws IOException {
        child.next();
    }
    
    public void prepareForWrite(boolean modified) throws IOException {
        child.prepareForWrite(modified);
    }
    
    public ISourceHandler<I, O> instanceFor(URI uriIn, URI uriOut) {
        throw new UnsupportedOperationException("Wrapper instance for other ISH objects. Created only explicitly.");
    }
    
    @Override
    public String toString() {
        return getClass().getSimpleName()+"("+getUri()+")";
    }
    
}
