/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import ssg.tools.easyconfig.utils.NFactoryPrototype;
import ssg.tools.easyconfig.data.NSettingSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.lang.reflect.Constructor;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.ServiceLoader;

/**
 *
 * @author ssg
 */
public class SourceHandlerFactory extends NFactoryPrototype<ISourceHandler> {

    private static ServiceLoader<ISourceHandler> handlersLoader = ServiceLoader.load(ISourceHandler.class);
    static SourceHandlerFactory factory;

    public static SourceHandlerFactory getInstance() {
        if (factory == null) {
            factory = new SourceHandlerFactory();
            factory.initDefaultTypes(true);
        }
        return factory;
    }

    @Override
    public ISourceHandler getType(String name, Object... parameters) {
        return super.getType(name, parameters);
    }

    public ISourceHandler newSourceHandler(URI uri) {
        List<ISourceHandler> handlers = findType(uri, false);
        if (handlers != null && !handlers.isEmpty()) {
            return handlers.get(0).instanceFor(uri, null);
        } else {
            for (ISourceHandler iss : handlersLoader) {
                try {
                    if (iss.canRead(uri)) {
                        return iss.instanceFor(uri, null);
                    }
                } catch (IOException ioex) {
                }
            }
        }
        return null;
    }

    // recognizes nested handlers if path contains "!".
    public ISourceHandler newSourceHandler(URI uri, NSettingSource nss) {
        if (nss != null && isAbsolutePath(nss.getPath())) {
            try {
                uri = new URI(nss.getPath());
            } catch (URISyntaxException usex) {
                usex.printStackTrace();
                return null;
            }
        } else if (nss != null && nss.getPath().contains("!")) {
            try {
                String p0 = uri.toString();
                String p1 = nss.getPath().substring(0, nss.getPath().lastIndexOf("!"));
                uri = new URI(uri.toString() + "/" + p1);
            } catch (URISyntaxException usex) {
                usex.printStackTrace();
                return null;
            }
        }
        return newSourceHandler(uri);
    }

    public boolean isAbsolutePath(String path) {
        return path != null && path.contains("//");
    }

    @Override
    public boolean matchType(ISourceHandler type, Object obj, boolean exact) {
        try {
            if (type.canRead(obj)) {
                if (obj instanceof InputStream) {
                    return true;
                }
                String s = obj.toString();
                if (s.contains(".")) {
                    s = s.substring(s.lastIndexOf(".") + 1);
                }
                String key = getOption("associations", s.toLowerCase());
                if (key == null || key.equalsIgnoreCase(s)) {
                    return true;
                } else if (key != null && type.equals(getType(key))) {
                    return true;
                } else if (!exact) {
                    return true;
                }
            }
        } catch (IOException ioex) {
            ioex.printStackTrace();
        }
        return false;
    }

    /*
     * Parses input stream as properties file with following named sections
     * format: [handlers] <type name>=<class name> [associations] <alias type
     * name>=<type name>
     *
     */
    @Override
    public void loadTypeDefinitions(InputStream is) throws IOException {
        LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is, "UTF-8"));
        String s = null;
        String section = "";

        for (ISourceHandler iss : handlersLoader) {
            //System.out.println("Available handler: " + iss);
        }

        while ((s = lnr.readLine()) != null) {
            s = s.trim();
            if (s.startsWith("#") || s.startsWith("//")) {
                continue;
            }
            if (s.startsWith("[")) {
                section = s.substring(1, s.length() - 1).toLowerCase();
                continue;
            }

            String key = s;
            String value = null;
            {
                int idx = key.indexOf("=");
                if (idx != -1) {
                    value = key.substring(idx + 1).trim();
                    key = key.substring(0, idx).trim();
                }
            }

            if (section.isEmpty() || "handlers".equalsIgnoreCase(section)) {
                if (value != null && !value.isEmpty()) {
                    boolean found = false;
                    for (ISourceHandler iss : handlersLoader) {
                        if (value.equals(iss.getClass().getName())) {
                            registerType(key, iss, true);
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        try {
                            Class cl = getClass().getClassLoader().loadClass(value);
                            Constructor c = cl.getConstructor(URI.class);
                            registerType(key, (ISourceHandler) c.newInstance((URI) null), true);
                        } catch (Throwable th) {
                            th.printStackTrace();
                        }
                    }
                }
            } else {
                if (value != null && !value.isEmpty()) {
                    String[] ss = value.split(",");
                    for (String a : ss) {
                        a = a.trim();
                        if (!a.isEmpty()) {
                            setOption(section, key, a);
                        }
                    }
                }
            }
        }
    }
}
