/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.handler.support;

import java.io.BufferedInputStream;
import ssg.tools.easyconfig.data.NSettingSource;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import ssg.tools.easyconfig.data.APPLY_STRATEGY;
import ssg.tools.easyconfig.utils.IOUtils;
import ssg.tools.easyconfig.utils.UncloseableInputStream;
import ssg.tools.easyconfig.utils.UncloseableOutputStream;

/**
 * Zip streams handler.
 * Supports operation in nested mode.
 *
 */
public class ZipStreamSourceHandler implements INestedSourceHandler<InputStream, OutputStream> {
    //

    APPLY_STRATEGY applyStrategy;
    // source
    URI uri;
    ISourceHandler<InputStream, OutputStream> parentSourceHandler;
    // reader
    ZipInputStream zis;
    ZipEntry ze;
    ZipEntry zeNext;
    // writer
    ZipOutputStream zos;
    OutputStream os;

    public ZipStreamSourceHandler() {
    }

    public ZipStreamSourceHandler(File file) {
        uri = file.toURI();
    }

    public ZipStreamSourceHandler(URI uri) {
        this.uri = uri;
    }

    public ZipStreamSourceHandler(ISourceHandler<InputStream, OutputStream> parent) {
        this.parentSourceHandler = parent;
    }

    /**
     * Initializes source baed on type. For read-only mode only input stream can
     * be requested and output stream is not accessible.
     *
     * @param readOnly
     * @throws IOException
     */
    @Override
    public void open(boolean readOnly, APPLY_STRATEGY applyStrategy) throws IOException {
        if (this.parentSourceHandler != null) {
            zis = new ZipInputStream(new UncloseableInputStream(this.parentSourceHandler.getInput()));
        } else if (uri != null) {
            zis = new ZipInputStream(uri.toURL().openStream());
        } else {
            throw new IOException("No zip source is available.");
        }
        if (!readOnly) {
            if (this.parentSourceHandler != null) {
                zos = new ZipOutputStream(new UncloseableOutputStream(this.parentSourceHandler.getOutput()));
            } else if ("file".equals(uri.getScheme())) {
                zos = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(new File((new File(uri)).getAbsolutePath() + ".tmp"))));
            } else {
                zos = null;
            }
        }
        this.applyStrategy = (applyStrategy != null) ? applyStrategy : APPLY_STRATEGY.replace;
    }

    /**
     * Returns true if next source item is available. Must be called before 1st
     * item is used.
     *
     * @return
     */
    @Override
    public boolean hasNext() {
        if (zis != null) {
            if (zeNext != null) {
                return true;
            }
            try {
                zeNext = zis.getNextEntry();
            } catch (IOException ioex) {
                // nothing more...
                ioex.printStackTrace();
            }
            return zeNext != null;
        } else {
            return false;
        }
    }

    /**
     * Makes next item current. Closes any previous opened resources (files,
     * streams, etc.)
     *
     * @throws IOException
     */
    @Override
    public void next() throws IOException {
        ze = zeNext;
        zeNext = null;
    }

    /**
     * Must be called before writing to output stream for current item. E.g.
     * prepares zip entry (see copyInputToOutput for details.
     *
     * @param modified
     * @throws IOException
     */
    @Override
    public void prepareForWrite(boolean modified) throws IOException {
        if (zos != null && ze != null) {
            if (modified) {
                ZipEntry ze2 = new ZipEntry(ze.getName());
                ze2.setComment(ze.getComment());
                ze2.setMethod(ze.getMethod());
                ze2.setExtra(ze.getExtra());
                ze = ze2;
            }
            zos.putNextEntry(ze);
        }
    }

    /**
     * Retruns current item name.
     *
     * @return
     */
    @Override
    public String getName() {
        return (ze != null) ? ze.getName() : null;
    }

    /**
     * Returns input stream for current item.
     *
     * @return
     * @throws IOException
     */
    @Override
    public InputStream getInput() throws IOException {
        return (ze != null) ? new FilterInputStream(zis) {

            @Override
            public void close() throws IOException {
                // do not allow to close stream
            }
        } : null;
    }

    /**
     * Returns output stream for current item to store modified version.
     *
     * @return
     * @throws IOException
     */
    @Override
    public <T extends OutputStream> T getOutput() throws IOException {
        return (T) zos;
    }

    @Override
    public void copyInputToOutput() throws IOException {
        prepareForWrite(false);
        if (zos != null) {
            byte[] buf = new byte[1024];
            InputStream eis = getInput();
            int count;
            while ((count = eis.read(buf)) != -1) {
                zos.write(buf, 0, count);
            }
        }
    }

    /**
     * Closes any open resources.
     *
     * @throws IOException
     */
    @Override
    public void close() throws IOException {
        // to close any open resources...
        next();
        boolean readOnly = zos == null;
        if (zis != null) {
            try {
                zis.close();
            } catch (IOException ioex) {
            }
            ze = null;
            if (zos != null) {
                zos.close();
                zos = null;
            }
        }

        if (!readOnly && uri != null && "file".equals(uri.getScheme())) {
            File oldF = new File(uri);
            File newF = new File(oldF.getAbsolutePath() + ".tmp");
            File bakF = new File(oldF.getAbsolutePath() + ".bak");
            switch (applyStrategy) {
                case replace:
                    oldF.delete();
                    newF.renameTo(oldF);
                    break;
                case backup:
                    if (bakF.exists()) {
                        int bakIdx = 0;
                        File bakF2 = new File(bakF.getAbsolutePath() + "." + bakIdx);
                        while (bakF2.exists()) {
                            bakIdx++;
                            bakF2 = new File(bakF.getAbsolutePath() + "." + bakIdx);
                        }
                        bakF = bakF2;
                    }
                    oldF.renameTo(bakF);
                    newF.renameTo(oldF);
                    break;
                case other:
                    break;
                default:
                    break;
            }
        }
    }
    static String[] defaultExtensions = new String[]{
        "zip", "jar", "war", "ear", "sar"
    };

    public boolean canRead(Object obj) throws IOException {
        if (obj instanceof File
                || obj instanceof URI
                || obj instanceof URL
                || obj instanceof String) {
            String s = obj.toString().toLowerCase();
            // ensure we have
            try {
                URI uri = IOUtils.string2uri(s);
                URL url = uri.toURL();
            } catch (Throwable th) {
                return false;
            }
            if (s.contains(".")) {
                s = s.substring(s.lastIndexOf(".") + 1);
            }
            for (String s2 : defaultExtensions) {
                if (s.equalsIgnoreCase(s2)) {
                    return true;
                }
            }
            // probe?
            InputStream is = null;
            try {
                is = new BufferedInputStream(IOUtils.uri2InputStream(IOUtils.string2uri("" + obj), null), 256);
                return canRead(is);
            } catch (Throwable th) {
                // we do ignore any problem when probing for archive...
            } finally {
                if (is != null) {
                    try {
                        is.close();
                    } catch (Throwable th2) {
                    }
                }
            }
        }
        if (obj instanceof NSettingSource) {
            NSettingSource nss = (NSettingSource) obj;
            if (nss.getPath().contains("//")) {
                return false;
            } else if (nss.getPath().contains("!")) {
                if (this.getUri().toString().contains("!")) {
                    String op1 = nss.getPath().substring(0, nss.getPath().lastIndexOf("!"));
                    String op2 = nss.getPath().substring(nss.getPath().lastIndexOf("!") + 1);
                    String p1 = getUri().toString().substring(0, getUri().toString().lastIndexOf("!"));
                    if (p1.endsWith(op1)) {
                        return true;
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            } else {
                return true;
            }
        }
        if (obj instanceof InputStream && ((InputStream) obj).markSupported()) {
            InputStream is = (InputStream) obj;
            is.mark(2);
            byte[] buf = new byte[2];
            is.read(buf);
            is.reset();
            return buf[0] == 'P' && buf[1] == 'K';
        }
        return false;
    }

    public void init(Object owner) throws IOException {
        // TODO: check if we need owner...
    }

    public URI getUri() {
        return uri;
    }

    public ISourceHandler<InputStream, OutputStream> instanceFor(URI uriIn, URI uriOut) {
        try {
            if (canRead(uriIn)) {
                return new ZipStreamSourceHandler(uriIn);
            }
        } catch (IOException ioex) {
        }
        return null;
    }

    public ISourceHandler<InputStream, OutputStream> getParentSource() {
        return this.parentSourceHandler;
    }

    public void setParentSource(ISourceHandler<InputStream, OutputStream> parent) {
        this.parentSourceHandler = parent;
    }
    
    @Override
    public String toString() {
        return getClass().getSimpleName()+"("+getUri()+")";
    }
}
