/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.sources;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.types.IBinarySettingType;
import ssg.tools.easyconfig.data.types.ISettingType;
import ssg.tools.easyconfig.data.types.NumericType;
import ssg.tools.easyconfig.data.types.SettingTypeFactory;

/**
 * Provides access to byte or sequence of bytes at given offset (that is setting
 * locator).
 * 
 * Only settings with type implementing IBinarySettingType are allowed (if not - setting is ignored).
 * Setting locator is interpreted as absolute offset in input stream (long value).
 * Overlapping settings are not supported.
 *
 * @author ssg
 */
public class ByteStreamSource extends NSettingSource implements IStreamSource {

    public static final String SOURCE_TYPE = "binary";

    public ByteStreamSource() {
        super(null, SOURCE_TYPE);
    }

    public ByteStreamSource(String path) {
        super(path, SOURCE_TYPE);
    }

    public boolean load(InputStream is) throws IOException {
        List<NSetting> settings = getPreparedSettings();
        Iterator<NSetting> it = settings.iterator();
        long pos = 0;
        while (it.hasNext()) {
            NSetting ns = it.next();
            long skip = NumericType.string2number(ns.getLocator(), 0L).longValue() - pos;
            if (skip > 0) {
                is.skip(skip);
                pos += skip;
                ISettingType st = SettingTypeFactory.getInstance().getType(ns.getType(), null);
                if (st instanceof IBinarySettingType) {
                    IBinarySettingType bst = (IBinarySettingType) st;
                    int oSize = bst.getBinarySize();
                    if (oSize > 0) {
                        byte[] data = new byte[oSize];
                        is.read(data);
                        pos += data.length;
                        try {
                            bst.setBinary(ns, data);
                            ns.setLoadedValue(ns.getValue());
                            ns.setLoadError(null);
                        } catch (Throwable th) {
                            ns.setLoadError(th);
                        }
                    }
                }
            }
        }
        while (is.read() != -1);
        return true;
    }

    public boolean save(InputStream is, OutputStream os) throws IOException {
        List<NSetting> settings = getPreparedSettings();
        Iterator<NSetting> it = settings.iterator();
        long pos = 0;
        while (it.hasNext()) {
            NSetting ns = it.next();
            long skip = NumericType.string2number(ns.getLocator(), 0L).longValue() - pos;
            if (skip > 0) {
                pos += skip;
                byte[] buf = new byte[1024];
                while (skip > 0) {
                    int c = is.read(buf, 0, (skip < buf.length) ? (int) skip : buf.length);
                    os.write(buf, 0, c);
                    skip -= c;
                }
                ISettingType st = SettingTypeFactory.getInstance().getType(ns.getType(), null);
                if (st instanceof IBinarySettingType) {
                    IBinarySettingType bst = (IBinarySettingType) st;
                    int oSize = bst.getBinarySize();
                    if (oSize > 0) {
                        byte[] data = new byte[oSize];
                        is.read(data);
                        pos += data.length;
                        if (ns.isModified()) {
                            byte[] nsdata = bst.getBinary(ns);
                            if (nsdata != null) {
                                try {
                                    os.write(nsdata);
                                    ns.setSaveError(null);
                                } catch (Throwable th) {
                                    os.write(data);
                                    ns.setSaveError(th);
                                }
                            }
                            ns.setLoadedValue(ns.getValue());
                        } else {
                            os.write(data);
                        }
                    }
                }
            }
            //
            byte[] buf = new byte[1024];
            int c = 0;
            while ((c = is.read(buf)) > 0) {
                os.write(buf, 0, c);
            }
        }
        return true;
    }

    @Override
    public NSettingSource instanceFor(String path) {
        return new ByteStreamSource(path);
    }

    /**
     * Filters out non IBinarySettingType type settings and orders settings by
     * offset (locator).
     * 
     * @return 
     */
    List<NSetting> getPreparedSettings() {
        List<NSetting> result = new ArrayList<NSetting>();
        result.addAll(getSettings());
        for (int i = result.size() - 1; i >= 0; i--) {
            ISettingType st = SettingTypeFactory.getInstance().getType(result.get(i).getType(), null);
            if (!(st instanceof IBinarySettingType)) {
                result.remove(i);
            }
        }
        Collections.sort(result, new Comparator<NSetting>() {

            public int compare(NSetting o1, NSetting o2) {
                long l1 = NumericType.string2number(o1.getLocator(), 0L).longValue();
                long l2 = NumericType.string2number(o2.getLocator(), 0L).longValue();
                return (l1 < l2) ? -1 : (l1 == l2) ? 0 : 1;
            }
        });
        return result;
    }
}
