/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.sources;

import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.data.NSetting;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Map;
import ssg.tools.easyconfig.utils.IOUtils;

/**
 * Groupped properties follow properties file format with extension of dividing properties into groups (sections).
 * Group is started with "[<group name>]" line and ends when next group is started.
 * Uninitialized group has name "[default]".
 * 
 * @author ssg
 */
public class GrouppedPropertiesSettingSource extends NSettingSource implements IWritableSource<InputStream, OutputStream> { //IStreamSource {

    public static final String SOURCE_TYPE = "groupped-properties";

    public GrouppedPropertiesSettingSource() {
        super(null, SOURCE_TYPE);
    }

    public GrouppedPropertiesSettingSource(String path) {
        super(path, SOURCE_TYPE);
    }

    @Override
    public boolean load(InputStream is) throws IOException {
        Map<String, Map<String, String>> groups = IOUtils.loadTextConfiguration(is, "UTF-8");

        for (NSetting ns : getSettings()) {
            String l = ns.getLocator();
            String g = (l.contains("].")) ? l.substring(1, l.indexOf("].")).toLowerCase() : "";
            String n = (l.contains("].")) ? l.substring(l.indexOf("].") + 2) : l;
            if (groups.containsKey(g)) {
                String value = groups.get(g).get(n);
                if (value != null) {
                    ns.setLoadedValue(value);
                    ns.setValue(value);
                    //System.out.println("   " + ns.getName() + " = " + ns.getLoadedValue());
                }
            }
        }
        setLoadedAt(System.currentTimeMillis());
        return true;
    }

    public boolean save(InputStream is, OutputStream os) throws IOException {
        LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is, "UTF-8"));
        Writer wr = new OutputStreamWriter(new BufferedOutputStream(os), "UTF-8");
        String group = "default";
        String s = null;
        while ((s = lnr.readLine()) != null) {
            if (!s.trim().startsWith("#") && !s.trim().startsWith("//")) {
                String separator = "";
                // test for group change
                if (s.startsWith("[")) {
                    group = s.substring(1, s.length() - 1).toLowerCase();
                } else {
                    // ensure valid '=' or ':' separator
                    int idx0 = s.indexOf(':');
                    int idx1 = s.indexOf('=');
                    if (idx0 > 1) {
                        if (s.charAt(idx0 - 1) == '\\') {
                            idx0 = -1;
                        }
                    }
                    if (idx1 > 1) {
                        if (s.charAt(idx1 - 1) == '\\') {
                            idx1 = -1;
                        }
                    }
                    if (idx0 != -1 && idx1 != -1) {
                        if (idx0 < idx1) {
                            idx1 = -1;
                        } else {
                            idx0 = -1;
                        }
                    }
                    if (idx0 != -1) {
                        separator = ":";
                    } else if (idx1 != -1) {
                        separator = "=";
                    }
                }
                if (!separator.isEmpty()) {
                    String[] ss = s.split(separator);
                    if (ss.length > 1) {
                        String key = ss[0].trim();
                        String value = (ss.length > 1) ? ss[1] : "";
                        for (NSetting ns : getSettings()) {
                            String l = ns.getLocator();
                            String g = (l.contains("].")) ? l.substring(1, l.indexOf("].")).toLowerCase() : "";
                            String n = (l.contains("].")) ? l.substring(l.indexOf("].") + 2) : l;
                            //System.out.println("   " + ns.getName() + " [ro=" + ns.isReadOnly() + ", mod=" + ns.isModified() + "] = " + ns.getValue() + " / " + ns.getLoadedValue());
                            if (g.equals(group) && n.equals(key) && ns.isModified() && !ns.isReadOnly()) {
                                s = key + separator + " " + ns.getValue();
                                //System.out.println("   " + ns.getName() + " = " + ns.getLoadedValue());
                                break;
                            }
                        }
                    }
                }
            }
            wr.write(s + "\n");
        }
        wr.flush();
        // loaded value = value -> no modifications left.
        for (NSetting ns : getSettings()) {
            ns.setLoadedValue(ns.getValue());
        }
        setAppliedAt(System.currentTimeMillis());
        return true;
    }

    @Override
    public NSettingSource instanceFor(String path) {
        if (path != null) {
            return new GrouppedPropertiesSettingSource(path);
        } else {
            return null;
        }
    }
}
