/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.sources;

import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NSettingSource;
import ssg.tools.easyconfig.utils.JDBCContext;
import java.io.IOException;
import java.net.URI;

/**
 * Provides read/write functionality for JDBC-based settings. Setting locator
 * syntax is: [<schema>.]<table[<where condition>]>.<column>
 *
 * @author ssg
 */
public class JDBCSettingSource extends NSettingSource implements IWritableSource<JDBCContext, JDBCContext> {

    public static final String SOURCE_TYPE = "jdbc";
    URI uri;

    public JDBCSettingSource() throws IOException {
        super(null, SOURCE_TYPE);
    }

    public JDBCSettingSource(String path) throws IOException {
        super(path, SOURCE_TYPE);
        init(path);
    }

    public JDBCSettingSource(URI uri) throws IOException {
        super((uri != null) ? uri.toString() : "", SOURCE_TYPE);
        init(uri.toString());
    }

    public boolean load(JDBCContext input) throws IOException {
        for (NSetting ns : getSettings()) {
            String value = input.read(ns.getLocator());
            if (value != null) {
                ns.setLoadedValue(value);
                ns.setValue(value);
            }
        }
        setLoadedAt(System.currentTimeMillis());
        return true;
    }

    public boolean save(JDBCContext input, JDBCContext output) throws IOException {
        for (NSetting ns : getSettings()) {
            if (!ns.isReadOnly()) {
                output.write(ns.getLocator(), ns.getValue());
            }
        }
        for (NSetting ns : getSettings()) {
            ns.setLoadedValue(ns.getValue());
        }
        setAppliedAt(System.currentTimeMillis());
        return true;
    }

    public void init(String jdbcURI) throws IOException {
        URI uri = string2JDBCUri(jdbcURI);
        if (uri != null) {
            this.uri = uri;
        } else {
            throw new IOException("Failed to initialize JDBC source for '" + jdbcURI + "': not valid jdbc URI.");
        }
    }

    URI string2JDBCUri(String jdbcURI) {
        try {
            URI uri = new URI(jdbcURI);
            String scheme = uri.getScheme();
            String primaryScheme = scheme;
            while (uri.getHost() == null) {
                uri = new URI(uri.getSchemeSpecificPart());
                if (uri.getScheme() == null) {
                    break;//throw new IOException("Invalid URI: " + jdbcURI);
                }
                scheme += ":" + uri.getScheme();
            }
            String auth = uri.getAuthority();
            String user = uri.getUserInfo();
            String host = uri.getHost();
            int port = uri.getPort();
            String path = uri.getPath();
            String query = uri.getQuery();
            String fragment = uri.getFragment();
            if ("jdbc".equals(primaryScheme)) {
                return new URI(jdbcURI);
            }
        } catch (Throwable th) {
            // TODO: ignore or not?!
        }
        return null;
    }

    @Override
    public NSettingSource instanceFor(String path) {
        URI uri = string2JDBCUri(path);
        if (uri != null) {
            try {
                return new JDBCSettingSource(uri);
            } catch (IOException ioex) {
                // TODO: log or ignore?!
            }
        }
        return null;
    }
}
