/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.types;

import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NValidationException;

/**
 *
 * @author ssg
 */
public class BooleanType extends BaseSettingType {

    static BooleanType extendedBoolean = new BooleanType(true, true);
    String[] trues = new String[]{"true"};
    String[] falses = new String[]{"false"};
    boolean allowNumeric = false;

    public BooleanType() {
        super(5);
    }

    public BooleanType(boolean extended, boolean allowNumeric) {
        super(5);
        if (extended) {
            trues = new String[]{"true", "yes", "on"};
            falses = new String[]{"false", "no", "off"};
        }
        this.allowNumeric = allowNumeric;
    }

    public BooleanType(String[] trues, String[] falses, boolean allowNumeric) {
        super(Math.max(maxSize(trues), maxSize(falses)));
        this.trues = trues;
        this.falses = falses;
        this.allowNumeric = allowNumeric;
    }

    public static int maxSize(String[] samples) {
        int result = 1;
        if (samples == null || samples.length == 0) {
            return result;
        }
        for (String s : samples) {
            if (s != null && s.length() > result) {
                result = s.length();
            }
        }
        return result;
    }

    public static boolean string2boolean(String value) throws NValidationException {
        return extendedBoolean.string2boolean(null, value);
    }

    public boolean string2boolean(NSetting ns, String value) throws NValidationException {
        if (falses != null) {
            for (String s : falses) {
                if (s != null && s.equalsIgnoreCase(value)) {
                    return false;
                }
            }
        }
        if (trues != null) {
            for (String s : trues) {
                if (s != null && s.equalsIgnoreCase(value)) {
                    return true;
                }
            }
        }
        if (allowNumeric) {
            try {
                Number n = NumericType.string2number(value, (value.toLowerCase().startsWith("0x") ? new Long(0) : null));
                return n.longValue() != 0;
            } catch (NumberFormatException nfex) {
                throw new NValidationException("Failed to interpret '" + value + "' as boolean");
            }
        }
        throw new NValidationException("Unknown boolean '" + value + "'");
    }

    public String validate(NSetting ns, String value) throws NValidationException {
        testMandatory(ns, value);
        if (value == null) {
            return null;
        }
        string2boolean(ns, value);
        return value;
    }

    public Class valueType() {
        return Boolean.class;
    }

    public Boolean asValue(String value) {
        if (value == null) {
            return null;
        }
        try {
            return string2boolean(value);
        } catch (NValidationException nvex) {
            return null;
        }
    }

    public boolean supports(String type) {
        return "boolean".equalsIgnoreCase(type) || "bool".equalsIgnoreCase(type);
    }

    public ISettingType instanceFor(String[] params) {
        if (params == null || params.length < 1 || !supports(params[0])) {
            return null;
        }
        return new BooleanType();
    }
}
