/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.data.types;

import ssg.tools.easyconfig.data.NSetting;
import ssg.tools.easyconfig.data.NValidationException;
import java.math.BigDecimal;
import java.math.BigInteger;
import ssg.tools.easyconfig.utils.ByteUtils;
import ssg.tools.easyconfig.utils.IOUtils;

/**
 *
 * @author ssg
 */
public class NumericType extends BaseSettingType implements IBinarySettingType {

    Number restriction;
    Number min;
    Number max;

    public NumericType() {
    }

    /**
     * Defines numeric validation with min/max values
     *
     * @param restriction
     * @param min
     * @param max
     */
    public NumericType(Number restriction, Number min, Number max) {
        super(type2size(restriction));
        this.restriction = restriction;
        this.min = min;
        this.max = max;
    }

    /**
     * Defines numeric validation with max string size
     *
     * @param restriction
     * @param size
     */
    public NumericType(Number restriction, int size) {
        super(size);
        this.restriction = restriction;
    }

    public NumericType(Number restriction) {
        super(type2size(restriction));
        this.restriction = restriction;
    }

    public static int type2size(Number type) {
        if (type == null) {
            type = Double.MAX_VALUE;
        }
        if (type instanceof Byte || byte.class.equals(type.getClass())) {
            return 4;
        } else if (type instanceof Short || short.class.equals(type.getClass())) {
            return 6;
        } else if (type instanceof Integer || int.class.equals(type.getClass())) {
            return 13;
        } else if (type instanceof Long || long.class.equals(type.getClass())) {
            return 25;
        } else {
            return 50;
        }
    }

    public static Number string2number(String value, Number likeThis) throws NumberFormatException {
        try {
            Number type = likeThis;
            Number n = null;
            if (type == null) {
                type = Double.MAX_VALUE;
            }
            int radix = 10;
            if (value.toLowerCase().startsWith("0x")) {
                radix = 16;
                value = value.substring(2);
            }
            if (type instanceof Byte || byte.class.equals(type.getClass())) {
                n = Byte.parseByte(value, radix);
            } else if (type instanceof Short || short.class.equals(type.getClass())) {
                n = Short.parseShort(value, radix);
            } else if (type instanceof Integer || int.class.equals(type.getClass())) {
                n = Integer.parseInt(value, radix);
            } else if (type instanceof Long || long.class.equals(type.getClass())) {
                n = Long.parseLong(value, radix);
            } else if (type instanceof BigInteger) {
                n = new BigInteger(value, radix);
            } else if (type instanceof BigDecimal) {
                if (radix == 10) {
                    n = new BigDecimal(value);
                } else {
                    n = new BigDecimal(new BigInteger(value, radix));
                }
            } else {
                if (radix == 10) {
                    if (type instanceof Float || float.class.equals(type.getClass())) {
                        n = Float.parseFloat(value);
                    } else if (type instanceof Double || double.class.equals(type.getClass())) {
                        n = Double.parseDouble(value);
                    }
                }
            }

            return n;
        } catch (Throwable th) {
            throw new NumberFormatException("Failed to convert '" + value + "' to number: " + th);
        }
    }

    public String validate(NSetting ns, String value) throws NValidationException {
        testMandatory(ns, value);
        try {
            Number type = restriction;
            Number n = null;
            if (type == null) {
                type = Double.MAX_VALUE;
            }
            int radix = 10;
            if (value.toLowerCase().startsWith("0x")) {
                radix = 16;
                value = value.substring(2);
            }
            if (type instanceof Byte || byte.class.equals(type.getClass())) {
                n = Byte.parseByte(value, radix);
            } else if (type instanceof Short || short.class.equals(type.getClass())) {
                n = Short.parseShort(value, radix);
            } else if (type instanceof Integer || int.class.equals(type.getClass())) {
                n = Integer.parseInt(value, radix);
            } else if (type instanceof Long || long.class.equals(type.getClass())) {
                n = Long.parseLong(value, radix);
            } else if (type instanceof BigInteger) {
                n = new BigInteger(value, radix);
            } else if (type instanceof BigDecimal) {
                if (radix == 10) {
                    n = new BigDecimal(value);
                } else {
                    n = new BigDecimal(new BigInteger(value, radix));
                }
            } else {
                if (radix == 10) {
                    if (type instanceof Float || float.class.equals(type.getClass())) {
                        n = Float.parseFloat(value);
                    } else if (type instanceof Double || double.class.equals(type.getClass())) {
                        n = Double.parseDouble(value);
                    }
                }
            }

            if (n != null) {
                if (min != null && min instanceof Comparable && ((Comparable) min).compareTo(n) < 0) {
                    throw new Exception("less than minimum (" + min + ")");
                }
                if (max != null && max instanceof Comparable && ((Comparable) max).compareTo(n) > 0) {
                    throw new Exception("more than maximum (" + max + ")");
                }
            }

            return value;
        } catch (Throwable th) {
            throw new NValidationException("Failed to interpret " + value + " as " + ((restriction != null) ? restriction.getClass().getName() : Double.class.getName()) + " number: " + th, th);
        }
    }

    public Class valueType() {
        return (restriction != null) ? restriction.getClass() : Double.class;
    }

    public Number asValue(String value) {
        try {
            return string2number(value, restriction);
        } catch (NumberFormatException nfex) {
            return null;
        }
    }

    public boolean supports(String type) {
        return "byte".equalsIgnoreCase(type)
                || "short".equalsIgnoreCase(type)
                || "int".equalsIgnoreCase(type)
                || "long".equalsIgnoreCase(type)
                || "float".equalsIgnoreCase(type)
                || "double".equalsIgnoreCase(type)
                || "decimal".equalsIgnoreCase(type)
                || "bigint".equalsIgnoreCase(type);
    }

    public ISettingType instanceFor(String[] params) {
        if (params == null || params.length < 1 || !supports(params[0])) {
            return null;
        }
        String tt = params[0];
        Number num = null;
        Number min = null;
        Number max = null;
        if ("byte".equalsIgnoreCase(tt)) {
            num = (byte) 0;
        } else if ("short".equalsIgnoreCase(tt)) {
            num = (short) 0;
        } else if ("int".equalsIgnoreCase(tt)) {
            num = (int) 0;
        } else if ("long".equalsIgnoreCase(tt)) {
            num = (long) 0;
        } else if ("float".equalsIgnoreCase(tt)) {
            num = 0f;
        } else if ("double".equalsIgnoreCase(tt)) {
            num = 0.0;
        } else if ("decimal".equalsIgnoreCase(tt)) {
            num = new BigDecimal(0);
        } else if ("bigint".equalsIgnoreCase(tt)) {
            num = new BigInteger("0");
        }
        // create numeric type...
        if (num != null) {
            String minS = IOUtils.findParam(params, "min");
            String maxS = IOUtils.findParam(params, "max");
            min = (minS != null) ? NumericType.string2number(minS, num) : null;
            max = (maxS != null) ? NumericType.string2number(maxS, num) : null;
            return new NumericType(num, min, max);
        } else {
            return null;
        }
    }

    ///////////////////////////////////////////////////////// IBinarySettingType
    public void setBinary(NSetting ns, byte[] data) throws NValidationException {
        ByteUtils bu = new ByteUtils(data);
        Number n = bu.getNumber(0, restriction);
        if (n != null) {
            ns.setValue(n.toString());
        } else {
            ns.setValue(null);
        }
    }

    public byte[] getBinary(NSetting ns) {
        int nSize = ByteUtils.sizeOf(restriction, null);
        if (nSize == -1) {
            return null;
        }
        ByteUtils bu = new ByteUtils(new byte[nSize]);
        bu.setNumber(0, (Number) asValue(ns.getValue()));
        return bu.getData();
    }

    public int getBinarySize() {
        return ByteUtils.sizeOf(restriction, null);
    }
}
