/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.Date;

/**
 *
 * @author ssg
 */
public class ByteUtils {

    byte[] data;
    int offset;

    public ByteUtils(byte[] data) {
        this.data = data;
    }

    public ByteUtils(byte[] data, int offset) {
        this.data = data;
        this.offset = offset;
    }

    public byte getByte(int offset) {
        return data[this.offset + offset];
    }

    public short getUByte(int offset) {
        return (short) (getByte(offset) & 0xFF);
    }

    public void setByte(int offset, byte b) {
        data[offset] = b;
    }

    public void setUByte(int offset, short b) {
        data[offset] = (byte) (b & 0xFF);
    }

    public long getLong(int offset) {
        return (long) ((getByte(offset + 7) & 0xFF) << 56
                | (getByte(offset + 6) & 0xFF) << 48
                | (getByte(offset + 5) & 0xFF) << 40
                | (getByte(offset + 4) & 0xFF) << 32
                | (getByte(offset + 3) & 0xFF) << 24
                | (getByte(offset + 2) & 0xFF) << 16
                | (getByte(offset + 1) & 0xFF) << 8
                | (getByte(offset) & 0xFF));
    }

    public void setLong(int offset, long l) {
        setByte(offset + 7, (byte) ((l >> 56) & 0xFF));
        setByte(offset + 6, (byte) ((l >> 48) & 0xFF));
        setByte(offset + 5, (byte) ((l >> 40) & 0xFF));
        setByte(offset + 4, (byte) ((l >> 32) & 0xFF));
        setByte(offset + 3, (byte) ((l >> 24) & 0xFF));
        setByte(offset + 2, (byte) ((l >> 16) & 0xFF));
        setByte(offset + 1, (byte) ((l >> 8) & 0xFF));
        setByte(offset + 0, (byte) (l & 0xFF));
    }

    public int getInt(int offset) {
        return ((getByte(offset + 3) & 0xFF) << 24
                | (getByte(offset + 2) & 0xFF) << 16
                | (getByte(offset + 1) & 0xFF) << 8
                | (getByte(offset) & 0xFF));
    }

    public long getUInt(int offset) {
        return (long) ((0 & 0xFF) << 56
                | (0 & 0xFF) << 48
                | (0 & 0xFF) << 40
                | (0 & 0xFF) << 32
                | (getByte(offset + 3) & 0xFF) << 24
                | (getByte(offset + 2) & 0xFF) << 16
                | (getByte(offset + 1) & 0xFF) << 8
                | (getByte(offset) & 0xFF));
    }

    public void setInt(int offset, int i) {
        setByte(offset + 3, (byte) ((i >> 24) & 0xFF));
        setByte(offset + 2, (byte) ((i >> 16) & 0xFF));
        setByte(offset + 1, (byte) ((i >> 8) & 0xFF));
        setByte(offset + 0, (byte) (i & 0xFF));
    }

    public void setUInt(int offset, long i) {
        setByte(offset + 3, (byte) ((i >> 24) & 0xFF));
        setByte(offset + 2, (byte) ((i >> 16) & 0xFF));
        setByte(offset + 1, (byte) ((i >> 8) & 0xFF));
        setByte(offset + 0, (byte) (i & 0xFF));
    }

    public short getShort(int offset) {
        return (short) ((getByte(offset + 1) & 0xFF) << 8
                | (getByte(offset) & 0xFF));
    }

    public int getUShort(int offset) {
        return ((0 & 0xFF) << 24
                | (0 & 0xFF) << 16
                | (getByte(offset + 1) & 0xFF) << 8
                | (getByte(offset) & 0xFF));
    }

    public void setShort(int offset, short sh) {
        setByte(offset + 1, (byte) ((sh >> 8) & 0xFF));
        setByte(offset + 0, (byte) (sh & 0xFF));
    }

    public void setUShort(int offset, int sh) {
        setByte(offset + 1, (byte) ((sh >> 8) & 0xFF));
        setByte(offset + 0, (byte) (sh & 0xFF));
    }

    public Number getNumber(int offset, Number type) {
        if (type instanceof Byte) {
            return getByte(offset);
        } else if (type instanceof Short) {
            return getShort(offset);
        } else if (type instanceof Integer) {
            return getInt(offset);
        } else if (type instanceof Long) {
            return getLong(offset);
        }
        return null;
    }

    public void setNumber(int offset, Number value) {
        if (value instanceof Byte) {
            setByte(offset, (Byte) value);
        } else if (value instanceof Short) {
            setShort(offset, (Short) value);
        } else if (value instanceof Integer) {
            setInt(offset, (Integer) value);
        } else if (value instanceof Long) {
            setLong(offset, (Long) value);
        }
    }

    public Date getDate(int offset) {
        return new Date(getLong(offset));
    }

    public Date getDate4(int offset) {
        return new Date(getInt(offset));

    }

    public String getString(int offset, int length, Charset charset) {
        return new String(data, this.offset + offset, length, charset);
    }

    /**
     * Puts specified string to data buffer using specified character set
     * encoding. Data size is limited by "maxLength" parameter. If "padding" is
     * not null then data if filled with the padding value up to "maxLength".
     *
     * Returns actual number of bytes used to store string value.
     *
     * @param s
     * @param offset
     * @param maxLength
     * @param charset
     * @param padding
     * @return
     */
    public int setString(String s, int offset, int maxLength, Charset charset, Byte padding) {
        int cnt = 0;
        if (s == null || s.isEmpty()) {
        } else {
            byte[] buf = s.getBytes(charset);
            cnt = Math.min(buf.length, maxLength);
            for (int i = 0; i < cnt; i++) {
                data[offset + i] = buf[i];
            }
        }
        if (padding != null) {
            for (int i = cnt; i < maxLength; i++) {
                data[i] = padding;
            }
        }
        return cnt;
    }

    public byte[] getData() {
        return data;
    }

    public static int sizeOf(Object obj, String encoding) {
        if (obj instanceof Byte) {
            return 1;
        } else if (obj instanceof Short) {
            return 2;
        } else if (obj instanceof Integer) {
            return 4;
        } else if (obj instanceof Long) {
            return 8;
        } else if (obj instanceof String) {
            if (encoding != null) {
                try {
                    return ((String) obj).getBytes(encoding).length;
                } catch (UnsupportedEncodingException ex) {
                    return -1;
                }
            } else {
                return ((String) obj).getBytes().length;
            }
        } else {
            return -1;
        }
    }
}
