/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 *
 * @author ssg
 */
public class IOUtils {

    public static List<String> listParamNames(String[] params) {
        List<String> result = new ArrayList<String>();
        if (params == null) {
            return result;
        }
        for (String s : params) {
            s = s.trim();
            int idx = s.indexOf("=");
            if (idx == -1) {
                continue;
            }
            String key = s.substring(0, idx).trim();
            result.add(key);
        }
        return result;
    }

    public static Map<String, Map<String, String>> loadTextConfiguration(InputStream is, String encoding) throws IOException {
        Map<String, Map<String, String>> sections = new LinkedHashMap<String, Map<String, String>>();
        LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is, (encoding != null) ? encoding : "UTF-8"));
        String s = null;
        String section = "";
        while ((s = lnr.readLine()) != null) {
            s = s.trim();
            if (s.startsWith("#") || s.startsWith("//")) {
                continue;
            }
            if (s.startsWith("[")) {
                section = s.substring(1, s.length() - 1).toLowerCase();
                continue;
            }
            String key = s;
            String value = null;
            {
                int idx = key.indexOf("=");
                if (idx != -1) {
                    value = key.substring(idx + 1).trim();
                    key = key.substring(0, idx).trim();
                }
            }
            if (section.isEmpty()) {
                section = "default";
            }
            if (value != null && !value.isEmpty()) {
                if (1 == 0) {
                    String[] ss = value.split(",");
                    for (String a : ss) {
                        a = a.trim();
                        if (!a.isEmpty()) {
                            Map<String, String> map = sections.get(section);
                            if (map == null) {
                                map = new LinkedHashMap<String, String>();
                                sections.put(section, map);
                            }
                            map.put(key, a);
                        }
                    }
                } else {
                    Map<String, String> map = sections.get(section);
                    if (map == null) {
                        map = new LinkedHashMap<String, String>();
                        sections.put(section, map);
                    }
                    map.put(key, value);
                }
            } else {
                Map<String, String> map = sections.get(section);
                if (map == null) {
                    map = new LinkedHashMap<String, String>();
                    sections.put(section, map);
                }
                map.put(key, "");
            }
        }
        return sections;
    }

    /**
     * Searches in string array for name=value and returns value or null if
     * none.
     *
     * @param params
     * @param name
     * @return
     */
    public static String findParam(String[] params, String name) {
        if (params == null || name == null) {
            return null;
        }
        for (String s : params) {
            s = s.trim();
            int idx = s.indexOf("=");
            if (idx == -1) {
                continue;
            }
            String key = s.substring(0, idx).trim();
            s = s.substring(idx + 1).trim();
            if (key.equalsIgnoreCase(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * Searches in string array for name=value and returns list of found values.
     *
     * @param params
     * @param name
     * @return
     */
    public static List<String> findParamValues(String[] params, String name) {
        List<String> result = new ArrayList<String>();
        if (params == null || name == null) {
            return result;
        }
        for (String s : params) {
            s = s.trim();
            int idx = s.indexOf("=");
            if (idx == -1) {
                continue;
            }
            String key = s.substring(0, idx).trim();
            s = s.substring(idx + 1).trim();
            if (key.equalsIgnoreCase(name)) {
                result.add(s);
            }
        }
        return result;
    }

    /**
     * Returns URI for string. First checks if URI is pointing to existing
     * file/folder, if not - tries to create URI object.
     *
     * @param path
     * @return
     */
    public static URI string2uri(String path) {
        if (path == null) {
            return null;
        }
        try {
            File f = new File(path);
            if (f.exists()) {
                return f.toURI();
            }
        } catch (Throwable th) {
        }
        try {
            return new URI(path);
        } catch (Throwable th1) {
            return null;
        }
    }

    /**
     * Tries to open input stream for object represented by URI.
     *
     * @param uri
     * @return
     * @throws IOException
     */
    public static InputStream uri2InputStream(URI uri, Map<String, Object> parameters) throws IOException {
        try {
            File f = new File(uri);
            return new FileInputStream(f);
        } catch (Throwable th) {
            URL url = uri.toURL();
            return url.openStream();
        }
    }

    /**
     * Tries to open output stream for object represented by URI.
     *
     * @param uri
     * @return
     * @throws IOException
     */
    public static OutputStream uri2OutputStream(URI uri, Map<String, Object> parameters) throws IOException {
        try {
            File f = new File(uri);
            return new FileOutputStream(f);
        } catch (Throwable th) {
            URL url = uri.toURL();
            // TODO: save settings template to URL?
            return null;
        }
    }

    ////////////////////////////////////////////////////////////// utilities
    /**
     * Recursive files lister.
     *
     * @param root
     * @param files
     */
    public static List<File> listFiles(File root, List<File> files) {
        if (files == null) {
            files = new ArrayList<File>();
        }
        if (root != null && root.exists() && root.isDirectory()) {
            File[] fs = root.listFiles();
            for (File f : fs) {
                files.add(f);
                if (f.isDirectory()) {
                    listFiles(f, files);
                }
            }
        }
        return files;
    }
}
