/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import ssg.tools.easyconfig.data.handler.support.SourceHandlerFactory;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Properties;

/**
 *
 * @author ssg
 */
public class JDBCContext {

    Connection connection;
    private URI uri;

    public JDBCContext() {
    }

    public JDBCContext(String uri) {
        try {
            this.uri = new URI(uri);
        } catch (URISyntaxException usex) {
            usex.printStackTrace();
        }
    }

    public JDBCContext(URI uri) {
        this.uri = uri;
    }

    protected Connection connect() throws SQLException {
        if (connection != null && !connection.isClosed()) {
            return connection;
        }
        try {
            String driverName = null;
            String dataSource = null;
            String user = null;
            String password = null;
            {
                String scheme = getUri().getScheme();
                while (getUri().getHost() == null) {
                    setUri(new URI(getUri().getSchemeSpecificPart()));
                    if (getUri().getScheme() == null) {
                        break; //throw new IOException("Invalid URI: " + jdbcURI);
                        //throw new IOException("Invalid URI: " + jdbcURI);
                    }
                    scheme += ":" + getUri().getScheme();
                }
                scheme += ":";
                String userInfo = getUri().getUserInfo();
                String host = getUri().getHost();
                int port = getUri().getPort();
                String path = getUri().getPath();
                String query = getUri().getQuery();
                String fragment = getUri().getFragment();

                driverName = SourceHandlerFactory.getInstance().getOption("jdbc", scheme);
                driverName = SourceHandlerFactory.getInstance().getOption("jdbc", driverName + ".driver");
                dataSource = scheme + "//" + host + ":" + port;
                if (path != null) {
                    dataSource += path;
                }
                if (query != null) {
                    dataSource += "?" + query;
                }
                if (fragment != null) {
                    dataSource += "#" + fragment;
                }
                if (userInfo != null) {
                    String[] ss = userInfo.split(":");
                    user = ss[0];
                    if (ss.length > 1) {
                        password = ss[1];
                    }
                }
            }
            Driver driver = (Driver) Class.forName(driverName).newInstance();
            Properties connectionProperties = new Properties();
            if (user != null) {
                connectionProperties.put("user", user);
                connectionProperties.put("username", user);
            }
            if (password != null) {
                connectionProperties.put("password", password);
            }
            connection = driver.connect(dataSource, connectionProperties);
            connection.setAutoCommit(false);
            //            metaData = connection.getMetaData();
            //            clientInfo = connection.getClientInfo();
            return connection;
        } catch (Exception e) {
            throw new SQLException("Problem opening DB connection: " + e.getMessage(), e);
        }
    }

    public void disconnect() throws SQLException {
        try {
            if (connection == null) {
                return;
            }
            if (!connection.isClosed()) {
                connection.close();
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new SQLException("Disconnection from DB failed: " + e.getMessage());
        }
        connection = null;
    }

    /**
     * @return the uri
     */
    public URI getUri() {
        return uri;
    }

    /**
     * @param uri the uri to set
     */
    public void setUri(URI uri) {
        this.uri = uri;
    }

    public String read(String expr) throws IOException {
        Expr pe = new Expr(expr);
        String sql = "select " + pe.column + " from " + pe.table + ((pe.where != null) ? " where " + pe.where : "");
        try {
            Connection conn = connect();
            PreparedStatement ps = conn.prepareStatement(sql);
            try {
                ResultSet rs = ps.executeQuery();
                if (rs.first()) {
                    String val = rs.getString(1);
                    rs.close();
                    return val;
                } else {
                    rs.close();
                }
            } finally {
                ps.close();
            }
        } catch (SQLException sex) {
            throw new IOException("Failed to read value with '" + expr + "': " + sex, sex);
        }
        return null;
    }

    public void write(String expr, Object value) throws IOException {
        Expr pe = new Expr(expr);
        String sql = "update " + pe.table + " set " + pe.column + "=? " + ((pe.where != null) ? " where " + pe.where : "");
        try {
            Connection conn = connect();
            PreparedStatement ps = conn.prepareStatement(sql);
            try {
                ps.setObject(1, value);
                int count = ps.executeUpdate();
                if (count == 0) {
                    throw new IOException("No data is updated with '" + expr + "'");
                }
            } finally {
                ps.close();
            }
        } catch (SQLException sex) {
            throw new IOException("Failed to write value with '" + expr + "': " + sex, sex);
        }
    }

    public class Expr {

        String table;
        String column;
        String where;

        public Expr(String expr) throws IOException {
            try {
                int idx = expr.indexOf('[');
                int idx2 = expr.lastIndexOf('.');
                column = expr.substring(idx2 + 1).trim();
                table = expr.substring(0, (idx != -1) ? idx : idx2);
                if (idx != -1) {
                    idx2 = expr.lastIndexOf(']');
                    where = expr.substring(idx + 1, idx2);
                }
            } catch (Throwable th) {
                throw new IOException("Failed to parse JDBC expression '" + expr + "': " + th, th);
            }
        }
    }
}
