/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * Utility factory class to unify factories usage accross application.
 *
 * @author ssg
 */
public abstract class NFactoryPrototype<T> {

    public static boolean DUMP_FACTORY_CONFIG = false;
    protected Map<String, T> types = new HashMap<String, T>();
    protected Map<String, Properties> options = new HashMap<String, Properties>();

    /**
     * Returns list of registered type "names". Factory may support others as
     * well but these listed are guaranteed.
     *
     * @return
     */
    public List<String> listRegisteredTypeNames() {
        List<String> result = new ArrayList<String>();
        result.addAll(types.keySet());
        return result;
    }

    /**
     * Registers type instance for type name optionally overriding old
     * registration. The registered type instance is not functional but is a
     * reference object.
     *
     * @param name
     * @param type
     * @param override
     * @return
     */
    public T registerType(String name, T type, boolean override) {
        if (name == null) {
            return null;
        }
        if (types.containsKey(name)) {
            if (override && type != null) {
                types.put(name, type);
            }
            return types.get(name);
        } else {
            if (type != null) {
                types.put(name, type);
            }
        }
        return type;
    }

    /**
     * Returns registered name type.
     *
     * @param name
     * @param parameters
     * @return
     */
    public T getType(String name, Object... parameters) {
        T result = types.get(name);
        return result;
    }

    /**
     * Returns list of setting types compatible with object. None is returned
     * for null and only only exact type (class) matching if exact = true;
     *
     * @param obj
     * @param exact
     * @return
     */
    public List<T> findType(Object obj, boolean exact) {
        List<T> result = new ArrayList();
        if (obj == null) {
            return result;
        }
        Class oc = obj.getClass();
        for (T t : types.values()) {
            if (matchType(t, obj, exact)) {
                result.add(t);
            }
        }
        return result;
    }

    /**
     * Returns true if specified type can handle specified object instance.
     * Since same object may be handled by several types the "exact" parameter
     * allows to control matching quality.
     *
     * This method must be implemented in actual type factory.
     *
     * @param type
     * @param obj
     * @param exact
     * @return
     */
    public abstract boolean matchType(T type, Object obj, boolean exact);

    /**
     * Removes type registration.
     *
     * @param typesToUnregister
     */
    public void unregisterTypes(Collection<String> typesToUnregister) {
        if (typesToUnregister == null) {
            return;
        }
        for (String type : typesToUnregister) {
            if (types.containsKey(type)) {
                types.remove(type);
            }
        }
    }

    /**
     * Loads type definitions (name - instance) and options from files
     * "META-INF/settings/<full class name>.properties"
     *
     * Settings syntax is sectioned properties file: i.e. section iss opened
     * with line "[<name>]" and all properties until next section are stored
     * under its name.
     *
     * Factory may decide which sections are loadable and which are ignored or
     * treated differently.
     *
     * @param reset
     */
    public void initDefaultTypes(boolean reset) {
        if (reset) {
            types.clear();
            options.clear();
        }
        try {
            //String rn = getClass().getPackage().getName().replace(".", "/") + "/" + getClass().getSimpleName() + ".properties";
            String rn = "META-INF/settings/" + getClass().getName() + ".properties";
            Enumeration<URL> urls = getClass().getClassLoader().getResources(rn);
            while (urls.hasMoreElements()) {
                loadTypeDefinitions(urls.nextElement().openStream());
            }
            //loadTypeDefinitions(this.getClass().getClassLoader().getResourceAsStream(rn));
        } catch (IOException ioex) {
            ioex.printStackTrace();
        }

        if (DUMP_FACTORY_CONFIG) {
            System.out.println("Factory " + getClass().getName() + " initialized as");
            System.out.println("  Types:");
            for (String key : types.keySet()) {
                if (types.get(key) != null) {
                    System.out.println("    " + key + " = " + types.get(key) + ((types.get(key).toString()!=null) ? " (" + types.get(key).getClass() + ")" : ""));
                } else {
                    System.out.println("    " + key + " = " + types.get(key));
                }
            }
            System.out.println("  Options:");
            for (String key : options.keySet()) {
                System.out.println("      " + key);
                Properties props = options.get(key);
                if (props != null) {
                    for (Object pn : Collections.list(props.propertyNames())) {
                        System.out.println("        " + pn + " = " + props.getProperty((String) pn));
                    }
                }
            }
        }
    }

    /**
     * Actually loads settings: types and options.
     *
     * @param is
     * @throws IOException
     */
    protected abstract void loadTypeDefinitions(InputStream is) throws IOException;
    ////////////////////////////////////////////////////////// utilities

    /**
     * Returns list of loaded option group names.
     *
     * @return
     */
    public Collection<String> getOptionGroupNames() {
        return options.keySet();
    }

    /**
     * Creates option group
     *
     * @param group
     * @return
     */
    public Properties createOptionGroup(String group) {
        if (group == null) {
            return null;
        } else {
            group = group.toLowerCase();
        }
        if (!options.containsKey(group)) {
            Properties props = new Properties();
            options.put(group, props);
        }
        return options.get(group);
    }

    /**
     * Returns list of option names for group.
     *
     * @param group
     * @return
     */
    public Collection<String> getOptionNames(String group) {
        if (group == null) {
            return null;
        } else {
            group = group.toLowerCase();
        }
        Properties props = options.get(group);
        if (props != null) {
            return (Collection<String>) Collections.list(props.propertyNames());
        } else {
            return Collections.emptyList();
        }
    }

    /**
     * Returns option value for group and option name.
     *
     * @param group
     * @param name
     * @return
     */
    public String getOption(String group, String name) {
        if (group == null) {
            return null;
        } else {
            group = group.toLowerCase();
        }
        Properties props = options.get(group);
        if (props != null) {
            return props.getProperty(name);
        } else {
            return null;
        }
    }

    /**
     * Assigns option value for option in group. If value is null then option is
     * removed from group.
     *
     * @param group
     * @param name
     * @param value
     */
    protected void setOption(String group, String name, String value) {
        if (group == null || name == null) {
            return;
        } else {
            group = group.toLowerCase();
        }
        Properties props = options.get(group);
        if (props == null) {
            props = createOptionGroup(group);
        }
        if (props != null) {
            if (value != null) {
                props.setProperty(name, value);
            } else if (props.getProperty(name) != null) {
                props.remove(name);
            }
        }
    }
}
