/*
 * Copyright 2012 sesidoro.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * "Scriptic" objects reader/modifier.
 * Uses variable parameters to specify path to data.
 * For setter last parameter is value to assign.
 * 
 * Provides static methods to retrieve some or all methods and fields for class.
 * Use options to control which methods/fields are of interest.
 * Default options include methods and fields, public only.
 * If options parameter is 0 then default options are used (DEFAULT_OPTIONS).
 * 
 * Use P.PP to share same options via instance.
 * 
 * NOTE: this class potentially overrides Java basic protection allowing access
 * to protected and private members and should be used with care and full understanding of
 * consequences.
 * 
 * NOTE: this class is copied from JSONXML library with exclusion of support of "Path" object.
 * 
 * @author ssg
 */
@SuppressWarnings("unchecked")
public class P {

    public static final int O_UNDEFINED = 0x0000;
    public static final int O_PUBLIC = 0x0001;
    public static final int O_PROTECTED = 0x0002;
    public static final int O_PRIVATE = 0x0004;
    public static final int O_FIELDS = 0x0008;
    public static final int O_METHODS = 0x0010;
    public static final int O_DO_NOT_SKIP_OVERRIDEN_FIELDS = 0x0020;
    public static final int DEFAULT_OPTIONS = O_PUBLIC | O_FIELDS | O_METHODS;
    // PP-handled expansions
    public static final int O_PATH_EXPAND = 0x0200;
    public static final int O_DOT_SEPARATED_STRING_EXPAND = 0x0400;
    public static final int O_SLASH_SEPARATED_STRING_EXPAND = 0x0800;
    public static final int O_OTHER_EXPAND = 0x1000;
    public static final int EXPANSIONS = O_PATH_EXPAND | O_DOT_SEPARATED_STRING_EXPAND | O_SLASH_SEPARATED_STRING_EXPAND | O_OTHER_EXPAND;

    /**
     * Evaluates property for given sequence of property names.
     * Interprets byte/short/integer/long as array/list index.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static Object get(Object object, int options, Object... property) {
        return read(object, options, 0, property);
    }

    /**
     * Returns "true" if value exists and is boolean true or if it is parsed as boolean true.
     * Otherwise returns "false".
     * 
     * @param object
     * @param property
     * @return 
     */
    public static Boolean getBool(Object object, int options, Object... property) {
        Object result = read(object, options, 0, property);
        if (result instanceof Boolean) {
            return (Boolean) result;
        }
        if (result instanceof String) {
            try {
                return Boolean.parseBoolean((String) result);
            } catch (Throwable th) {
            }
        }
        return false;
    }

    /**
     * Tries to return value as List.
     * Returns null if value is not present or throws cast exception if not list.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static List getList(Object object, int options, Object... property) {
        return (List) read(object, options, 0, property);
    }

    /**
     * Tries to return value as List.
     * Returns null if value is not present or if not list.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static List getIfList(Object object, int options, Object... property) {
        Object o = read(object, options, 0, property);
        if (o instanceof List) {
            return (List) o;
        } else {
            return null;
        }
    }

    /**
     * Tries to return value as Map.
     * Returns null if value is not present or throws cast exception if not map.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static Map getMap(Object object, int options, Object... property) {
        return (Map) read(object, options, 0, property);
    }

    /**
     * Tries to return value as Map.
     * Returns null if value is not present or if not map.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static Map getIfMap(Object object, int options, Object... property) {
        Object o = read(object, options, 0, property);
        if (o instanceof Map) {
            return (Map) o;
        } else {
            return null;
        }
    }

    /**
     * Returns "toString()" value of object or null if object not found.
     * @param object
     * @param property
     * @return 
     */
    public static String getStr(Object object, int options, Object... property) {
        Object o = read(object, options, 0, property);
        return (o != null) ? o.toString() : null;
    }

    /**
     * Tests if object exists (i.e. not is null).
     * 
     * @param object
     * @param property
     * @return 
     */
    public static boolean exists(Object object, int options, Object... property) {
        return read(object, options, 0, property) != null;
    }

    /**
     * Tests if object is missing (i.e. is null).
     * 
     * @param object
     * @param property
     * @return 
     */
    public static boolean missing(Object object, int options, Object... property) {
        return read(object, options, 0, property) == null;
    }

    /**
     * Evaluates properties in sequence resolving porperty as:
     * - key if currently evaluated object is Map
     * - int index if currently evaluated object is List or Array
     * - method name or getter method name or, if ends with "()" - method interpreted as getter.
     * - field name if no matching method was found
     * 
     * @param object
     * @param ignore - # of parameters to ignore ate the end of parameters list. 2 is used in "set", 1 - in "delete".
     * @param property
     * @return 
     */
    private static Object read(Object object, int options, int ignore, Object... property) {
        Object result = object;
        if (object == null || property == null) {
            return null;
        }
        if (options == O_UNDEFINED) {
            options = DEFAULT_OPTIONS;
        }
        //System.out.println(property.getClass() + ": " + property);
        int maxIdx = ((Object[]) property).length;
        if (ignore > 0 && ignore <= maxIdx) {
            maxIdx -= ignore;
        }
        for (Object p : property) {
            if (maxIdx <= 0) {
                continue;
            }
            maxIdx--;
            if (result == null) {
                continue;
            }
            boolean explicitMethod = (p instanceof String && ((String) p).endsWith("()"));
            if (result instanceof Map && !explicitMethod) {
                result = ((Map) result).get(p);
                continue;
            } else if (p instanceof String) {
                boolean resolved = false;
                Collection<Method> ms = methods(result.getClass(), options, null); // result.getClass().getMethods();
                if (ms != null) {
                    String mn = (String) p;
                    if (mn.endsWith("()")) {
                        mn = mn.substring(0, mn.length() - 2);
                    } else {
                        mn = "get" + mn.substring(0, 1).toUpperCase() + mn.substring(1);
                    }
                    for (Method m : ms) {
                        if ((m.getName().equals(mn) || m.getName().equals(p)) && (m.getParameterTypes() == null || m.getParameterTypes().length == 0)) {
                            try {
                                result = m.invoke(result);
                                resolved = true;
                                break;
                            } catch (Throwable th) {
                                System.out.println("Failed to call " + mn + ": " + th);
                                return null;
                            }
                        }
                    }
                }
                if (resolved) {
                    continue;
                }
                Collection<Field> fs = fields(result.getClass(), options, null); //result.getClass().getFields();
                if (fs != null) {
                    for (Field f : fs) {
                        if (f.getName().equals(p)) {
                            try {
                                result = f.get(result);
                                resolved = true;
                                break;
                            } catch (Throwable th) {
                                System.out.println("Failed to get field " + p + ": " + th);
                                return null;
                            }
                        }
                    }
                }
                if (resolved) {
                    continue;
                }
                return null;
            } else {
                try {
                    int i = 0 + ((Number) p).intValue();
                    if (result instanceof List) {
                        result = ((List) result).get(i);
                    } else if (result.getClass().getComponentType() != null) {
                        result = ((Object[]) result)[i];
                    }
                } catch (Throwable th) {
                    System.out.println("Failed to get list/array item " + p + ": " + th);
                    return null;
                }
            }
        }
        return result;
    }

    /**
     * Assigns value to object's property.
     * Properties list must be no less that 2 items. Last item is value to assign.
     * Uses "prop" to get nearest object for setting value (skipping 2 last parameters: property and value).
     * Resolves last property name as key for map, index for list(array), or field/setter for other object
     * and assigns value.
     * 
     * @param object
     * @param property 
     * @return true if assignment succeeded
     */
    public static boolean set(Object object, int options, Object... property) {
        if (object == null || property == null) {
            return false;
        }
        int pCount = ((Object[]) property).length;
        if (pCount < 2) {
            return false;
        }
        if (options == O_UNDEFINED) {
            options = DEFAULT_OPTIONS;
        }
        object = read(object, options, 2, property);
        if (object != null) {
            Object lastP = ((Object[]) property)[pCount - 2];
            Object value = ((Object[]) property)[pCount - 1];

            if (object instanceof Map) {
                ((Map) object).put(lastP, value);
                return true;
            } else if (object instanceof List) {
                List l = (List) object;
                int idx = 0 + ((Number) lastP).intValue();
                if (idx == l.size()) {
                    l.add(value);
                    return true;
                } else {
                    l.set(idx, value);
                    return true;
                }
            } else if (object.getClass().isArray()) {
                Object[] a = (Object[]) object;
                int idx = 0 + ((Number) lastP).intValue();
                a[idx] = value;
                return true;
            } else {
                // reflection
                Collection<Method> ms = methods(object.getClass(), options, null); //object.getClass().getMethods();
                if (ms != null) {
                    String mn = "" + lastP;
                    mn = "set" + mn.substring(0, 1).toUpperCase() + mn.substring(1);
                    for (Method m : ms) {
                        if (m.getName().equals(mn) && (m.getParameterTypes() == null || m.getParameterTypes().length == 1)) {
                            try {
                                m.invoke(object, value);
                                return true;
                            } catch (Throwable th) {
                                // we ignore this trial assuming there may be alternative setter...
                            }
                        }
                    }
                }

                Collection<Field> fs = fields(object.getClass(), options, null); //object.getClass().getFields();
                if (fs != null) {
                    for (Field f : fs) {
                        if (f.getName().equals(lastP)) {
                            try {
                                f.set(object, value);
                                return true;
                            } catch (Throwable th) {
                                //System.out.println("Failed to get field " + lastP + ": " + th);
                                return false;
                            }
                        }
                    }
                }
            }
        } else {
            // ignore
        }
        return false;
    }

    /**
     * Adds last parameter to property of type Collection (e.g. Set or List instance),
     * identified by previous "property" parameters.
     * 
     * @param object
     * @param property
     * @return 
     */
    public static boolean add(Object object, int options, Object... property) {
        if (object == null || property == null) {
            return false;
        }
        int pCount = ((Object[]) property).length;
        if (pCount < 1) {
            return false;
        }
        if (options == O_UNDEFINED) {
            options = DEFAULT_OPTIONS;
        }
        object = read(object, options, 1, property);
        if (object != null) {
            Object value = ((Object[]) property)[pCount - 1];

            if (object instanceof Collection) {
                Collection c = (Collection) object;
                c.add(value);
                return true;
            }
        } else {
            // ignore
        }
        return false;
    }

    /**
     * Assigns value to object's property.
     * Properties list must be no less that 2 items. Last item is value to assign.
     * Uses "prop" to get nearest object for setting value (skipping 2 last parameters: property and value).
     * Resolves last property name as key for map, index for list(array), or field/setter for other object
     * and assigns value.
     * 
     * @param object
     * @param property 
     * @return true if deletion succeeded
     */
    public static boolean delete(Object object, int options, Object... property) {
        if (object == null || property == null) {
            return false;
        }
        int pCount = ((Object[]) property).length;
        if (pCount < 1) {
            return false;
        }
        if (options == O_UNDEFINED) {
            options = DEFAULT_OPTIONS;
        }
        object = read(object, options, 1, property);
        if (object instanceof Collection || object instanceof Map) {
            Object idx = ((Object[]) property)[pCount - 1];

            if (object instanceof Map && ((Map) object).containsKey(idx)) {
                ((Map) object).remove(idx);
                return true;
            } else if (object instanceof List && idx instanceof Number) {
                Number num = (Number) idx;
                if (num.intValue() < ((List) object).size() && num.intValue() >= 0) {
                    ((List) object).remove(num.intValue());
                    return true;
                }
            } else if (object instanceof Set && ((Set) object).contains(idx)) {
                ((Set) object).remove(idx);
                return true;
            } else if (object instanceof Collection && ((Collection) object).contains(idx)) {
                ((Collection) object).remove(idx);
                return true;
            }
        } else if (object != null) {
            // TODO: do we have any other deletes but from collection or map?
        }
        return false;
    }
    private static Map<String, List<Method>> _cacheMethods = new HashMap<String, List<Method>>();
    private static Map<String, List<Field>> _cacheFields = new HashMap<String, List<Field>>();

    /**
     * Returns list of all class methods according to options.
     * 
     * @param clazz
     * @param options
     * @param result
     * @return 
     */
    public static synchronized Collection<Method> methods(Class clazz, int options, Map<String, Method> cache) {
        Collection<Method> result = null;
        if (cache == null) {
            cache = new LinkedHashMap<String, Method>();
        }
        if (_cacheMethods.containsKey(clazz.getName() + "_" + options)) {
            result = new ArrayList<Method>();
            result.addAll(_cacheMethods.get(clazz.getName() + "_" + options));
        } else {
            if ((options & (O_PROTECTED | O_PRIVATE)) == 0) {
                result = new ArrayList<Method>();
                result.addAll(Arrays.asList(clazz.getMethods()));
            } else {
                boolean includeProtected = (options & O_PROTECTED) != 0;
                boolean includePrivate = (options & O_PRIVATE) != 0;
                Method[] ms = clazz.getDeclaredMethods();
                for (Method m : ms) {
                    if (cache.containsKey(m.getName())) {
                        continue;
                    }
                    if ((m.getModifiers() & Modifier.PRIVATE) != 0 && includePrivate) {
                        m.setAccessible(true);
                        cache.put(m.getName(), m);
                    } else if (((m.getModifiers() & Modifier.PROTECTED) != 0 || m.getModifiers() == 0) && includeProtected) {
                        m.setAccessible(true);
                        cache.put(m.getName(), m);
                    } else if ((m.getModifiers() & Modifier.PUBLIC) != 0) {
                        cache.put(m.getName(), m);
                    }
                }
                if (!Object.class.equals(clazz.getSuperclass())) {
                    methods(clazz.getSuperclass(), options, cache);
                }
            }
            if (result == null) {
                result = cache.values();
            }
            List<Method> lst = new ArrayList<Method>(result.size());
            lst.addAll(result);
            _cacheMethods.put(clazz.getName() + "_" + options, lst);
        }
        return result;
    }

    /**
     * Returns list of all class fields according to options.
     * If options include O_DO_NOT_SKIP_OVERRIDEN_FIELDS and private or protected members are
     * also requested, then result will include also possibly overriden fields in parent class(es).
     * 
     * @param clazz
     * @param options
     * @param result
     * @return 
     */
    public static synchronized Collection<Field> fields(Class clazz, int options, Map<String, Field> cache) {
        Collection<Field> result = null;
        if (cache == null) {
            cache = new LinkedHashMap<String, Field>();
        }
        if (_cacheFields.containsKey(clazz.getName() + "_" + options)) {
            result = new ArrayList<Field>();
            result.addAll(_cacheFields.get(clazz.getName() + "_" + options));
        } else {
            if ((options & (O_PROTECTED | O_PRIVATE)) == 0) {
                result = new ArrayList<Field>();
                result.addAll(Arrays.asList(clazz.getFields()));
            } else {
                boolean includeProtected = (options & O_PROTECTED) != 0;
                boolean includePrivate = (options & O_PRIVATE) != 0;
                Field[] ms = clazz.getDeclaredFields();
                for (Field m : ms) {
                    if (cache.containsKey(m.getName())) {
                        if ((options & O_DO_NOT_SKIP_OVERRIDEN_FIELDS) == 0) {
                            continue;
                        }
                    }
                    if ((m.getModifiers() & Modifier.PRIVATE) != 0 && includePrivate) {
                        m.setAccessible(true);
                        cache.put(m.getName(), m);
                    } else if (((m.getModifiers() & Modifier.PROTECTED) != 0 || m.getModifiers() == 0) && includeProtected) {
                        m.setAccessible(true);
                        cache.put(m.getName(), m);
                    } else if ((m.getModifiers() & Modifier.PUBLIC) != 0) {
                        cache.put(m.getName(), m);
                    }
                }
                if (!Object.class.equals(clazz.getSuperclass())) {
                    fields(clazz.getSuperclass(), options, cache);
                }
            }
            if (result == null) {
                result = cache.values();
            }
            List<Field> lst = new ArrayList<Field>(result.size());
            lst.addAll(result);
            _cacheFields.put(clazz.getName() + "_" + options, lst);
        }
        return result;
    }

    /**
     * Extensions to P with alternative options and extended interpretation of parameters.
     */
    public static class PP {

        int _options = DEFAULT_OPTIONS;

        public PP() {
        }

        public PP(int options) {
            _options = options;
        }

        public void setOptions(int options) {
            _options |= options;
        }

        public void unsetOptions(int options) {
            if (hasOptions(options, false)) {
                _options ^= _options & options;
            }
        }

        public boolean hasOptions(int options, boolean any) {
            if (options == O_UNDEFINED) {
                return false;
            }
            return (any) ? (_options & options) != 0 : (_options & options) == options;
        }

        public Object get(Object object, Object... property) {
            return P.get(object, _options, toParameters(property));
        }

        public Boolean getBool(Object object, Object... property) {
            return P.getBool(object, _options, toParameters(property));
        }

        public List getList(Object object, Object... property) {
            return P.getList(object, _options, toParameters(property));
        }

        public List getIfList(Object object, Object... property) {
            return P.getIfList(object, _options, toParameters(property));
        }

        public Map getMap(Object object, Object... property) {
            return P.getMap(object, _options, toParameters(property));
        }

        public Map getIfMap(Object object, Object... property) {
            return P.getIfMap(object, _options, toParameters(property));
        }

        public String getStr(Object object, Object... property) {
            return P.getStr(object, _options, toParameters(property));
        }

        public boolean exists(Object object, Object... property) {
            return P.exists(object, _options, property);
        }

        public boolean missing(Object object, Object... property) {
            return P.missing(object, _options, property);
        }

        public boolean set(Object object, Object... property) {
            return P.set(object, _options, toParameters(property, true));
        }

        public boolean add(Object object, Object... property) {
            return P.add(object, _options, toParameters(property, true));
        }

        public boolean delete(Object object, Object... property) {
            return P.delete(object, _options, toParameters(property));
        }

        public Object[] toParameters(Object[] properties) {
            return toParameters(properties, false);
        }

        public Object[] toParameters(Object[] properties, boolean lastIsValue) {
            if (properties == null) {
                return properties;
            }

            List result = new ArrayList();
            int SKIP_IDX = properties.length - 1;
            boolean CAN_EXPAND = hasOptions(EXPANSIONS, true);
            if (!lastIsValue) {
                SKIP_IDX++;
            }
            for (int i = 0; i < properties.length; i++) {
                Object o = properties[i];
                if (CAN_EXPAND && i != SKIP_IDX) {
                    if (hasOptions(O_SLASH_SEPARATED_STRING_EXPAND, false) && o instanceof String && ((String) o).contains("/")) {
                        String[] oos = ((String) o).split("/");
                        for (String s : oos) {
                            if (s.length() > 0) {
                                if (s.contains("[")) {
                                    try {
                                        int ooi = Integer.parseInt(s.substring(s.indexOf("[") + 1, s.lastIndexOf("]")));
                                        result.add(s.substring(0, s.indexOf("[")));
                                        result.add(ooi);
                                    } catch (Throwable th) {
                                        result.add(s);
                                    }
                                } else {
                                    result.add(s);
                                }
                            }
                        }
                        continue;
                    } else if (hasOptions(O_DOT_SEPARATED_STRING_EXPAND, false) && o instanceof String && ((String) o).contains(".")) {
                        String[] oos = ((String) o).split("\\.");
                        for (String s : oos) {
                            if (s.length() > 0) {
                                if (s.length() > 0) {
                                    if (s.contains("[")) {
                                        try {
                                            int ooi = Integer.parseInt(s.substring(s.indexOf("[") + 1, s.lastIndexOf("]")));
                                            result.add(s.substring(0, s.indexOf("[")));
                                            result.add(ooi);
                                        } catch (Throwable th) {
                                            result.add(s);
                                        }
                                    } else {
                                        result.add(s);
                                    }
                                }
                            }
                        }
                        continue;
                    } else if (hasOptions(O_OTHER_EXPAND, false)) {
                        if (customExpand(o, result)) {
                            continue;
                        }
                    }
                }
                result.add(o);
            }
            return result.toArray();
        }

        public boolean customExpand(Object object, List result) {
            return false;
        }
    }
}
