/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * @author ssg
 */
public class Tree<T> {

    public static int MAX_DEPTH = 1000;
    T item;
    Tree<T> parent;
    Set<Tree<T>> children = new LinkedHashSet<Tree<T>>();

    public Tree(T item) {
        this.item = item;
    }

    public Tree(T item, T child) {
        this.item = item;
        addChild(child);
    }

    public Tree(T item, Collection<T> children) {
        this.item = item;
        if (children != null) {
            for (T child : children) {
                addChild(child);
            }
        }
    }

    public Tree(T item, T[] children) {
        this.item = item;
        if (children != null) {
            for (T child : children) {
                addChild(child);
            }
        }
    }

    public void addChild(T child) {
        Tree<T> tt = new Tree(child);
        if (child != null && !children.contains(tt)) {
            tt.parent = this;
            children.add(tt);
        }
    }

    public void removeChild(T child) {
        if (child != null && children.contains(child)) {
            children.remove(child);
        }
    }

    public Collection<T> getChildren() {
        List<T> result = new ArrayList<T>(children.size());
        for (Tree<T> ch : children) {
            result.add(ch.item);
        }
        return result;
    }

    public boolean isChild(T item, int level) {
        Tree<T> tt = find(item, level);
        return tt != null;
    }

    public boolean hasChildren() {
        return !children.isEmpty();
    }

    public Tree<T> find(T child, int level) {
        if (item != null && item.equals(child)) {
            return this;
        }
        if (hasChildren()) {
            for (Tree<T> tt : children) {
                return tt.find(child, level - 1);
            }
        }
        return null;
    }

    public List<T> path(T child) {
        Tree<T> tt = find(child, MAX_DEPTH);
        if (tt == null) {
            return Collections.singletonList(child);
        } else {
            List<T> result = new ArrayList<T>();
            while (tt != null) {
                result.add(0, tt.item);
                tt = tt.parent;
            }
            return result;
        }
    }

    @Override
    public String toString() {
        return item + ((!children.isEmpty()) ? " [" + children.size() + "] " + children : "");
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Tree<T> other = (Tree<T>) obj;
        if (this.item != other.item && (this.item == null || !this.item.equals(other.item))) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 11 * hash + (this.item != null ? this.item.hashCode() : 0);
        return hash;
    }
    
}
