/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils.tar;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

/**
 *
 * @author ssg
 */
/**
 * Stream implementation used to read data in TAR blocks from source stream
 * no more than provided size (with padding to TAR block size).
 */
public class NestedTarInputStream extends InputStream {
    byte[] buf = new byte[0];
    int idx = 0;
    InputStream sourceStream;
    String sourceName;
    long sizeLeft;
    long localCount = 0;

    public NestedTarInputStream(InputStream is, long size) {
        sourceStream = is;
        sizeLeft = size;
    }

    @Override
    public boolean markSupported() {
        return false;
    }

    @Override
    public long skip(long n) throws IOException {
        byte[] buf = new byte[1024];
        long skipped = 0;
        while (n > 0) {
            int cnt = read(buf, 0, Math.min(buf.length, (n > Integer.MAX_VALUE) ? Integer.MAX_VALUE : (int) n));
            if (cnt > 0) {
                n -= cnt;
                skipped += cnt;
            } else {
                break;
            }
        }
        sizeLeft -= skipped;
        localCount += skipped;
        return skipped;
    }

    @Override
    public int read() throws IOException {
        if (idx < buf.length) {
            localCount++;
            return buf[idx++] & 255;
        } else {
            if (sizeLeft > 0 && nextBlock() > 0) {
                localCount++;
                return buf[idx++] & 255;
            } else {
                return -1;
            }
        }
    }

    synchronized int nextBlock() throws IOException {
        if (sizeLeft > 0) {
            TarRecord tr = new TarRecord(sourceStream);
            buf = Arrays.copyOfRange(tr.data, 0, tr.getLength());
            sizeLeft -= buf.length;
            idx = 0;
            return buf.length;
        } else {
            buf = new byte[0];
            idx = 0;
            return -1;
        }
    }

    @Override
    public int available() throws IOException {
        return (getSizeLeft() > Integer.MAX_VALUE) ? Integer.MAX_VALUE : (int) getSizeLeft();
    }

    public long getSizeLeft() {
        return sizeLeft;
    }
    
}
