/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils.tar;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Date;
import java.util.zip.GZIPInputStream;

/**
 * This is stream-oriented implementation of Tar archives reader. It scans TAR
 * input stream and allows unpacking and/or decompression of files into
 * optionally separate output stream. For each file it creatres context and this
 * context is used to check if file may/should be expanded (as TAR archive or as
 * gzipped file).
 *
 * It uses globally defined properties for allowing expansion of nested TAR
 * archives and for decompressions ("expansion") of gzipped files.
 *
 * To control which files/archives are expanded just override context-specific
 * methods "isExpandableArchive()" and "isExpandableFile()".
 *
 * To process tar archive use "read()" method. This method returns read
 * statistics. To process files content override "createOutputStream()" and
 * "closeOutputStream()" methods. The former is used to obtain stream to put
 * file content. If null is returned then file content is not processed (i.e.
 * skipped).
 *
 * NOTE: most constants are taken from spec but never used, sorry.
 *
 * @author ssg
 */
public class Tar {
    public boolean logEnabled = false;
    
    public void log(String s) {
        if (logEnabled) {
            System.out.println(s);
        }
    }

    /**
     * @return the expandArchives
     */
    public boolean isExpandArchives() {
        return expandArchives;
    }

    /**
     * @param expandArchives the expandArchives to set
     */
    public void setExpandArchives(boolean expandArchives) {
        this.expandArchives = expandArchives;
    }

    /**
     * @return the expandFiles
     */
    public boolean isExpandFiles() {
        return expandFiles;
    }

    /**
     * @param expandFiles the expandFiles to set
     */
    public void setExpandFiles(boolean expandFiles) {
        this.expandFiles = expandFiles;
    }
    
    public boolean isExpandableArchive(ReadContext ctx) {
        return isExpandArchives() && ctx.isTar();
    }
    
    public boolean isExpandableFile(ReadContext ctx) {
        return ctx.isCompressed() && isExpandFiles();
    }
    
    public static class ReadStatistics {
        
        int folders;
        int files;
        int others;
        int expandedArchives;
        int expandedFiles;
        long size;
        long uncompressedSize;
        
        @Override
        public String toString() {
            return "Dirs=" + folders + ", files=" + files + ", others=" + others + ", expanded=" + expandedArchives + ", uncompressed=" + expandedFiles + ": size uncompressed=" + uncompressedSize + ", packed=" + size;
        }
        
        public void update(TarEntryHeader h) {
            if (h != null && h.getSize() != -1) {
                switch (h.getMagic()) {
                    case TarEntryHeader.LF_OLDNORMAL:
                    case TarEntryHeader.LF_NORMAL:
                        files++;
                        break;
                    case TarEntryHeader.LF_DIR:
                        folders++;
                        break;
                    case TarEntryHeader.LF_LINK:
                    case TarEntryHeader.LF_SYMLINK:
                    case TarEntryHeader.LF_CHR:
                    case TarEntryHeader.LF_BLK:
                    case TarEntryHeader.LF_FIFO:
                    case TarEntryHeader.LF_CONTIG:
                    default:
                        others++;
                }
            }
        }
        
        public void update(long size, long uncompressedSize) {
            if (size < 0 || uncompressedSize < 0) {
                return;
            }
            this.size += size;
            this.uncompressedSize += uncompressedSize;
            if (uncompressedSize != size) {
                expandedFiles++;
            }
        }
        
        public void update(long size, ReadStatistics expanded) {
            if (expanded == null) {
                return;
            }
            folders += expanded.folders;
            files += expanded.files - 1; // exclude self as file
            others += expanded.others;
            expandedArchives += expanded.expandedArchives + 1; // for this expansion
            expandedFiles += expanded.expandedFiles;
            this.size += size;
            uncompressedSize += expanded.uncompressedSize;
        }
    }

    /**
     * Keeps info on current file/directory
     */
    public static class ReadContext {
        
        public static final long O_COMPRESSED = 0x10000000;
        public static final long O_TAR = 0x20000000;
        ReadContext parent;
        String path = "";
        String name = "";
        long options = 0L;
        private Date time;
        long size;
        private long actualSize;
        
        public ReadContext(File file) {
            if (file == null) {
                // no file ?
            } else if (file.exists()) {
                if (file.isDirectory()) {
                    path = file.getAbsolutePath() + "/";
                } else {
                    path = file.getParent() + "/";
                    name = file.getName();
                    size = file.length();
                    actualSize = size;
                }
            } else {
                // assume it is directory
                path = file.getAbsolutePath() + "/";
            }
            evaluateOptions();
        }
        
        public ReadContext(ReadContext parent, TarEntryHeader h) {
            this.parent = parent;
            if (h != null && h.getSize() != -1) {
                String s = "";//h.getFilenamePrefix();
                s += ((s.isEmpty()) ? "" : "/") + h.getName();
                if (s.endsWith("/")) {
                    path = s;
                } else {
                    if (s.contains("/")) {
                        path = s.substring(0, s.lastIndexOf("/") + 1);
                        name = s.substring(s.lastIndexOf("/") + 1);
                        size = h.getSize();
                        time = h.getMTime();
                    } else {
                        name = s;
                        size = h.getSize();
                        time = h.getMTime();
                    }
                }
            }
            evaluateOptions();
        }
        
        public void evaluateOptions() {
            if (!name.isEmpty()) {
                String lc = name.toLowerCase();
                if (lc.endsWith(".gz") || lc.endsWith(".tgz")) {
                    options |= O_COMPRESSED;
                }
                if (lc.endsWith(".tar") || lc.endsWith(".tgz") || lc.endsWith(".tar.gz")) {
                    options |= O_TAR;
                }
            }
        }
        
        public String getRootPath() {
            return (parent == null) ? path : parent.getRootPath();
        }
        
        public String getLocalPath() {
            return path;
        }
        
        public String getFullPath() {
            return ((parent != null) ? parent.getFullPath() : "") + getLocalPath();
        }
        
        public String getName(boolean uncompressed) {
            if (uncompressed && isCompressed()) {
                String lc = name.toLowerCase();
                if (lc.endsWith(".gz")) {
                    return name.substring(0, name.length() - 3);
                } else if (lc.endsWith(".tgz")) {
                    return name.substring(0, name.length() - 4);
                } else {
                    return name;
                }
            } else {
                return name;
            }
        }
        
        public String getLocalName(boolean uncompressed) {
            return getLocalPath() + getName(uncompressed);
        }
        
        public String getFullName(boolean uncompressed) {
            return getFullPath() + getName(uncompressed);
        }
        
        public long getSize() {
            return size;
        }
        
        public int getDepth() {
            return (parent != null) ? parent.getDepth() + 1 : 0;
        }
        
        public boolean isDirectory() {
            return !path.isEmpty() && name.isEmpty();
        }
        
        public boolean isFile() {
            return !name.isEmpty();
        }
        
        public boolean isCompressed() {
            return (options & O_COMPRESSED) == O_COMPRESSED;
        }
        
        public boolean isTar() {
            return (options & O_TAR) == O_TAR;
        }

        /**
         * @return the time
         */
        public Date getTime() {
            return time;
        }

        /**
         * @return the actualSize
         */
        public long getActualSize() {
            return actualSize;
        }
    }
    private boolean expandArchives = false;
    private boolean expandFiles = false;
    
    public ReadStatistics read(ReadContext parentCtx, InputStream is) throws IOException {
        ReadStatistics result = new ReadStatistics();
        int cnt = 0;
        while (true) {
            TarEntryHeader h = new TarEntryHeader(is);
            long fileSize = h.getSize();
            if (fileSize == -1) {
                log(parentCtx.getFullPath() + ": EOF: " + h);
                break;
            }
            h.testChkSum();
            result.update(h);
            
            ReadContext ctx = new ReadContext(parentCtx, h);
            boolean isNested = false;

            //System.out.println(ctx.getLocalPath() + ": [" + cnt + "] (size=" + fileSize + "): " + h);

            if (fileSize > 0) {
                if (isExpandableArchive(ctx)) {
                    isNested = true;
                    log("........... enter nested for " + ctx.getLocalName(false));
                    NestedTarInputStream nestedTarStream = new NestedTarInputStream(is, fileSize);
                    nestedTarStream.sourceName = ctx.getLocalName(ctx.isCompressed() && isExpandArchives());
                    InputStream nis = (ctx.isCompressed()) ? new GZIPInputStream(nestedTarStream) : nestedTarStream;
                    ReadStatistics expandResult = read(ctx, nis);
                    result.update(h.getSize(), expandResult);
                    fileSize = nestedTarStream.getSizeLeft();
                    log("........... exit  nested for " + ctx.getLocalName(ctx.isCompressed() && isExpandArchives()));
                }
                if (fileSize > 0) {
                    // just skip if no OS is defined or write to OS
                    OutputStream os = null;
                    if (!isNested) {
                        os = createOutputStream(ctx);
                    }
                    InputStream nis = null;
                    long localCount = 0;
                    try {
                        int countBlocks = 0;
                        final NestedTarInputStream nestedTarStream = new NestedTarInputStream(is, fileSize);
                        nestedTarStream.sourceName = ctx.getLocalName(ctx.isCompressed() && isExpandFiles());
                        boolean gzipped = isExpandableFile(ctx) && !isNested;
                        nis = (gzipped) ? new GZIPInputStream(nestedTarStream) : nestedTarStream;
                        TarRecord r = null;
                        while ((r = new TarRecord(nis)).getLength() >= 0 || fileSize > 0) {
                            int rSize = (gzipped) ? r.getLength() : (fileSize >= 0 && fileSize < r.getLength()) ? (int) fileSize : r.getLength();
                            if (os != null && rSize > 0) {
                                os.write(r.data, 0, rSize);
                                log("read (" + nestedTarStream.sourceName + ", gz=" + gzipped + "): todo=" + nestedTarStream.getSizeLeft() + ", loaded total=" + localCount + ", last loaded=" + r.getLength());
                            }
                            if (rSize > 0) {
                                localCount += rSize;
                            }
                            fileSize = (gzipped) ? nestedTarStream.getSizeLeft() : fileSize - rSize;
                            countBlocks++;
                        }
                    } finally {
                        if (!isNested) {
                            ctx.actualSize = localCount;
                            closeOutputStream(ctx, os);
                        }
                    }
                    result.update(h.getSize(), localCount);
                    
                    if (nis != null) {
                        nis.close();
                    }
                }
            } else if (fileSize == 0 && h.getMagic() == TarEntryHeader.LF_OLDNORMAL || h.getMagic() == TarEntryHeader.LF_NORMAL) {
                // create (or try to create) empty file
                OutputStream os = createOutputStream(ctx);
                closeOutputStream(ctx, os);
            }
            cnt++;
        }
        return result;
    }

    /**
     * Returnes opened OutputStream to write file content or null if data is not
     * needed.
     *
     * @param ctx
     * @return
     * @throws IOException
     */
    public OutputStream createOutputStream(ReadContext ctx) throws IOException {
        String fileName = ctx.getFullName(ctx.isCompressed() && isExpandFiles());
        File file = new File(fileName);
        file.getParentFile().mkdirs();
        return new BufferedOutputStream(new FileOutputStream(file));
    }

    /**
     * Closes output stream opened by
     * <code>createOutputStream</code> method.
     *
     * @param ctx
     * @param os
     * @throws IOException
     */
    public void closeOutputStream(ReadContext ctx, OutputStream os) throws IOException {
        if (os != null) {
            os.flush();
            os.close();
        }
    }
}
