/*
 * Copyright 2012 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ssg.tools.easyconfig.utils.tar;

import java.io.IOException;
import java.io.InputStream;
import java.util.Date;

/**
 *
 * @author ssg
 */
public class TarEntryHeader {
    /*
     * Standard Archive Format - Standard TAR - USTAR
     */

    public static final int NAMSIZ = 100;
    public static final int TUNMLEN = 32;
    public static final int TGNMLEN = 32;
    /*
     * The checksum field is filled with this while the checksum is computed.
     */
    public static final String CHKBLANKS = "        "; // 8 blanks, no null
    /*
     * The magic field is filled with this if this is a GNU format dump entry
     */
    public static final String GNUMAGIC = "GNUtar "; // 7 chars and a null
    /*
     * The magic field is filled with this if uname and gname are valid.
     */
    public static final String TMAGIC = "ustar  "; // 7 chars and a null
    /*
     * The linkflag defines the type of file
     */
    public static final char LF_OLDNORMAL = (char) 0; // Normal disk file, Unix compatible
    public static final char LF_NORMAL = '0'; // Normal disk file
    public static final char LF_LINK = '1'; // Link to previously dumped file
    public static final char LF_SYMLINK = '2'; // Symbolic link
    public static final char LF_CHR = '3'; // Character special file
    public static final char LF_BLK = '4'; // Block special file
    public static final char LF_DIR = '5'; // Directory
    public static final char LF_FIFO = '6'; // FIFO special file
    public static final char LF_CONTIG = '7'; // Contiguous file
    /*
     * Further link types may be defined later.
     */
    /*
     * Bits used in the mode field - values in octal
     */
    public static final int TSUID = 2048; // Set UID on execution
    public static final int TSGID = 1024; // Set GID on execution
    public static final int TSVTX = 512; // Save text (sticky bit)
    /*
     * File permissions
     */
    public static final int TUREAD = 256; // read by owner
    public static final int TUWRITE = 128; // write by owner
    public static final int TUEXEC = 64; // execute/search by owner
    public static final int TGREAD = 32; // read by group
    public static final int TGWRITE = 16; // write by group
    public static final int TGEXEC = 8; // execute/search by group
    public static final int TOREAD = 4; // read by other
    public static final int TOWRITE = 2; // write by other
    public static final int TOEXEC = 1; // execute/search by other
    //
    public static final int SZ = 100 + 8 + 8 + 8 + 12 + 12 + 8 + 1 + 100 + 8 + 32 + 32 + 8 + 8 + 155;
    byte[] data = new byte[512];
    int length = 0;

    public TarEntryHeader(InputStream is) throws IOException {
        int cnt = is.read(data);
        // Iteratively ensure all header bytes are read.
        // This is needed, e.g., when reading from GZipInputStream.
        while (cnt < data.length) {
            int lc = is.read(data, cnt, data.length - cnt);
            if (lc < 0) {
                throw new IOException("EOF detected: tar entry header loading is unfinished[" + cnt + " out of 512 bytes]: " + this.toString());
            }
            cnt += lc;
        }
        length = cnt;
    }

    public String getName() {
        return (new String(data, 0, NAMSIZ)).trim();
    }

    public void setName(String name) {
        // TODO: check tar file name encoding limitations
        setString(name, 0, NAMSIZ);
    }

    public String getMode() {
        return (new String(data, NAMSIZ, 8)).trim();
    }

    public void setMode(String mode) {
        setString(mode, NAMSIZ, 8);
    }

    public String getUid() {
        return (new String(data, NAMSIZ + 8, 8)).trim();
    }

    public void setUid(String uid) {
        setString(uid, NAMSIZ + 8, 8);
    }

    public String getGid() {
        return (new String(data, NAMSIZ + 8 + 8, 8)).trim();
    }

    public void setGid(String gid) {
        setString(gid, NAMSIZ + 8 + 8, 8);
    }

    public long getSize() {
        try {
            return Long.parseLong((new String(data, NAMSIZ + 8 + 8 + 8, 12)).trim(), 8);
        } catch (Throwable th) {
        }
        return -1;
    }

    public void setSize(long size) {
        setString(Long.toOctalString(size), NAMSIZ + 8 + 8 + 8, 12);
    }

    public Date getMTime() {
        try {
            return new Date(1000 * Long.parseLong((new String(data, NAMSIZ + 8 + 8 + 8 + 12, 12)).trim(), 8));
        } catch (Throwable th) {
            return null;
        }
    }

    public void setMTime(Date date) {
        setString(Long.toOctalString((date!=null) ? date.getTime() / 1000 : 0), NAMSIZ + 8 + 8 + 8 + 12, 12);
    }

    public String getChkSum() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12, 8)).trim();
    }

    public char getMagic() {
        return (char) (data[NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8] & 255);
    }

    public String getUName() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1, TUNMLEN)).trim();
    }

    public String getGName() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN, TGNMLEN)).trim();
    }

    public String getDevMajor() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN + TGNMLEN, 8)).trim();
    }

    public String getDevMinor() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN + TGNMLEN + 8, 8)).trim();
    }

    public String getFilenamePrefix() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN + TGNMLEN + 8 + 8, 155)).trim();
    }

    public String getUndefined() {
        return (new String(data, NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN + TGNMLEN + 8 + 8 + 155, data.length - (NAMSIZ + 8 + 8 + 8 + 12 + 12 + 8 + 1 + TUNMLEN + TGNMLEN + 8 + 8 + 155))).trim();
    }

    public boolean testChkSum() {
        int idx = NAMSIZ + 8 + 8 + 8 + 12 + 12;
        int sum = 0;
        for (int i = 0; i < idx; i++) {
            sum += (data[i] & 255);
        }
        sum += 32 * 8;
        for (int i = idx + 8; i < data.length; i++) {
            sum += (data[i] & 255);
        }
        int refSum = Integer.parseInt(getChkSum(), 8);
        if (sum != refSum) {
            System.out.println("Header checksum=" + (sum == refSum) + ": " + sum + "/" + refSum + " (" + getChkSum() + ")");
        }
        return sum == refSum;
    }

    private int setString(String s, int offset, int max) {
        return setBytes(s.getBytes(), offset, max);
    }

    private int setBytes(byte[] buf, int offset, int max) {
        for (int i = 0; i < Math.min(buf.length, max); i++) {
            data[offset + i] = buf[i];
        }
        if (buf.length < max) {
            for (int i = buf.length; i < max; i++) {
                data[i] = ' ';
            }
        }
        return max;
    }

    @Override
    public String toString() {
        return getName() + " [t=" + getMagic() + ", m=" + getMode() + ", uid/gid=" + getUid() + "/" + getGid() + ", sz=" + getSize() + ", dt=" + getMTime() + ", chk=" + getChkSum() + ", u/g=" + getUName() + "/" + getGName() + ", dev=" + getDevMajor() + "/" + getDevMinor() + ", prefix=" + getFilenamePrefix() + ", und=" + getUndefined() + "]";
    }
}
